# https://github.com/pyg-team/pytorch_geometric/blob/72eb1b38f60124d4d700060a56f7aa9a4adb7bb0/torch_geometric/nn/models/pg_explainer.py

import math
import torch.nn as nn
import torch
import copy
import numpy as np
import pandas as pd
from scipy.special import softmax
from pgmpy.estimators.CITests import chi_square
from torch_geometric.data import Data, Batch
from ..base import BaseRandom
from tqdm import tqdm
# print = tqdm.write
# from evaluation import control_sparsity


class MetaPGMExplainer:
    def __init__(
            self,
            model,
            graph,
            pred_threshold=0.5,
            perturb_mode="mean",  # mean, zero, max or uniform
            perturb_indicator="diff",
    ):
        self.model = model
        # self.model.eval()
        self.graph = graph
        # self.num_layers = num_layers
        # self.perturb_feature_list = perturb_feature_list
        self.perturb_mode = perturb_mode
        self.perturb_indicator = perturb_indicator
        # self.perturb_pct = perturb_pct
        self.X_feat = graph.x
        self.geo_feat = graph.pos
        self.pred_threshold = pred_threshold

    def perturb_on_graph(self, node_idx, x_level):
        candi_feat = self.X_feat if x_level == 'graph' else self.geo_feat if x_level == 'geometric' else None

        # X_perturb = copy.deepcopy(candi_feat)
        perturb_array = copy.deepcopy(candi_feat)
        epsilon = 0.05 * torch.max(candi_feat, dim=0).values
        epsilon = epsilon.detach().cpu().numpy()

        for i in node_idx:
            if self.perturb_mode == "mean":
                perturb_array[i] = torch.mean(candi_feat, dim=0)
            elif self.perturb_mode == "zero":
                perturb_array[i] = torch.zeros((candi_feat.shape[1], 1))
            elif self.perturb_mode == "max":
                perturb_array[i] = torch.max(candi_feat, dim=0).values
            elif self.perturb_mode == "uniform":
                perturb_array[i] = perturb_array[i] + np.random.uniform(low=-epsilon, high=epsilon)
            else:
                assert self.perturb_mode == 'split'
                tmp_graph = self.graph.clone().cpu()
                all_idx = np.arange(self.graph.num_nodes)
                node_idx = np.setdiff1d(all_idx, node_idx)

                new_x = tmp_graph.x[node_idx]
                new_pos = None if tmp_graph.pos==None else tmp_graph.pos[node_idx]

                edge_index = tmp_graph.edge_index
                row = edge_index[0]
                new_idx = row.new_full((self.graph.num_nodes + 1,), -1)
                # print(new_idx.device)
                new_idx[node_idx] = torch.arange(node_idx.size)
                new_edge_index = new_idx[edge_index]

                new_edge_index = new_edge_index.T[(new_edge_index.T != torch.tensor([-1, -1])).all(dim=1)].T
                # delete the -1 ones

                tmp_graph.x = new_x
                tmp_graph.pos = new_pos
                tmp_graph.edge_index = new_edge_index

                return tmp_graph.to(self.graph.x.device)

        tmp_graph = self.graph.clone()
        if x_level == 'graph':
            tmp_graph.x = perturb_array
        elif x_level == 'geometric':
            tmp_graph.pos = perturb_array
        else:
            raise ValueError(f'Unknown x_level: {x_level}')

        return tmp_graph

    def batch_perturb_features_on_node(self, num_samples, index_to_perturb,
                                       percentage, x_level):
        clf_logits = self.model(self.graph)
        soft_pred = clf_logits.sigmoid().detach().cpu().numpy()
        num_nodes = self.X_feat.size(0)
        Samples = []
        for iteration in range(num_samples):
            # perturb_feat = copy.deepcopy(candi_feat)
            sample = np.zeros((num_nodes+1,))

            perturb_num = int(percentage / 100 * num_nodes)

            perturb_indexes = np.random.permutation(index_to_perturb)[:perturb_num]
            sample[perturb_indexes] = 1
            perturb_graph = self.perturb_on_graph(perturb_indexes, x_level)

            perturb_logits = self.model(perturb_graph)
            soft_pred_perturb = perturb_logits.sigmoid().detach().cpu().numpy()
            # self.model(tmp_g)
            # soft_pred_perturb = np.asarray(softmax(self.model.readout[0].detach().cpu().numpy()))

            pred_change = (soft_pred - soft_pred_perturb).item()
            sample[-1] = pred_change
            Samples.append(sample)
        Samples = np.asarray(Samples)
        # np.set_printoptions(precision=0, suppress=True, threshold=np.inf)
        if self.perturb_indicator == "abs":
            Samples = np.abs(Samples)
        #
        # for i in range(num_samples):
        count = 0
        reverse_val = 0
        for i in range(num_samples):
            if self.pred_threshold:
                if Samples[i, num_nodes] > self.pred_threshold:
                    count += 1
                    reverse_val += Samples[i, num_nodes]
                    Samples[i, num_nodes] = 1
                else:
                    Samples[i, num_nodes] = 0
            else:  # select most significant 1/8 changes and set to 1
                top = int(num_samples / 8)
                top_idx = np.argsort(Samples[:, num_nodes])[-top:]
                if i in top_idx:
                    count += 1
                    reverse_val += Samples[i, num_nodes]
                    Samples[i, num_nodes] = 1
                else:
                    Samples[i, num_nodes] = 0
        # if self.graph.y.item() == 1:
        #     print(f'The flipped ratio is {(count/num_samples):.2f}, average drop is {reverse_val}/{count}.')
        # if reverse_val == 0:
        #     print(Samples[:, num_nodes])
        return Samples

    def explain(self, x_level, num_samples=1000, percentage=20, top_node=5, p_threshold=0.05):

        num_nodes = self.X_feat.size(0)

        #       Round 1
        Samples = self.batch_perturb_features_on_node(int(num_samples / 2), range(num_nodes), percentage, x_level)

        data = pd.DataFrame(Samples)
        p_values = []

        target = num_nodes  # The entry for the graph classification result is at "num_nodes"
        for node in range(num_nodes):
            chi2, p, _ = chi_square(node, target, [], data, boolean=False)
            p_values.append(p)

        number_candidates = min(int(top_node * 2), num_nodes - 1)
        candidate_nodes = np.argpartition(p_values, number_candidates)[0:number_candidates]

        #         Round 2
        Samples = self.batch_perturb_features_on_node(num_samples, candidate_nodes, percentage, x_level)
        data = pd.DataFrame(Samples)

        p_values = []
        dependent_nodes = []

        target = num_nodes
        for node in range(num_nodes):
            chi2, p, _ = chi_square(node, target, [], data, boolean=False)
            p_values.append(p)
            if p < p_threshold:
                dependent_nodes.append(node)
        return torch.tensor(p_values, device=self.graph.x.device)


class PGMExplainer(BaseRandom):

    def __init__(self, clf, criterion, config):
        super().__init__()
        self.name = 'pgmexplainer'
        self.clf = clf
        self.device = next(self.parameters()).device
        # self.extractor = extractor
        # extractor wasn't used
        self.criterion = criterion
        self.pred_threshold = config['pred_threshold']
        self.percentage = config['percentage']
        self.perturb_mode = config['perturb_mode']

    def forward_pass(self, data, epoch, do_sampling):
        x_level = 'geometric'
        clf_logits = self.clf(data)
        batch_imp = []
        for graph in data.to_data_list():
            node_imp = self.explain_graph(graph, x_level)
            batch_imp += [node_imp]
        # imp = self.explain_graph(data, x_level)
        res_weights = self.min_max_scalar(torch.cat(batch_imp))
        return -1, {}, clf_logits, res_weights

    def explain_graph(self, graph, x_level):

        clf_logits = self.clf(graph)
        soft_pred = clf_logits.sigmoid()
        # self.model(graph)
        # soft_pred = self.model.readout

        # pred_threshold = 0.1 * torch.max(soft_pred)
        # perturb_features_list = [i ]
        explainer = MetaPGMExplainer(self.clf, graph,
                                     perturb_indicator="abs",
                                     perturb_mode=self.perturb_mode,
                                     pred_threshold=self.pred_threshold)
        pct = self.percentage / 100
        # int(graph.num_nodes / pct)
        p_values = explainer.explain(x_level=x_level, num_samples=500, p_threshold=0.05,
                                     percentage=self.percentage, top_node=int(pct * graph.num_nodes))
        # p_values = np.array(p_values)
        row, col = graph.edge_index #.detach().cpu()
        edge_imp = (1-p_values[row]) * (1-p_values[col])
        edge_imp -= torch.min(edge_imp)

        if isinstance(edge_imp, float):
            edge_imp = edge_imp.reshape([1])

        def norm_imp(imp):
            imp[imp < 0] = 0
            imp += 1e-16
            return imp / imp.sum()
        edge_imp = norm_imp(edge_imp)

        # edge_imp if x_level == 'graph' else
        return 1-p_values if x_level == 'geometric' else edge_imp


EPS = 1e-6





import pandas as pd
import numpy as np
from pathlib import Path

# 初始化一个列表来存储每个文件的test_clf_auc值
def test():
    test_precision_values = []


    def collect_files_in_path(path_name):
        path = Path(path_name)
        


    # 遍历文件名
    for i in range(10):  # 假设文件名从0到9
        file_name = f"/nethome/jzhu617/renaissance/result/pointtrans_plbind/lri_gaussian/bs{i}_lri_gaussian_{i}.csv"  # 构造文件名
        df = pd.read_csv(file_name)  # 读取CSV文件
        # 假设每个文件只有一行数据，使用iloc[0]来获取这行数据
        # 如果有多行，可以根据具体情况调整
        test_precision = df['test_precision@20'].iloc[0]  
        test_precision_values.append(test_precision)  # 将值添加到列表中

    # 计算平均值和标准差
    avg_test_precision = np.mean(test_precision_values)
    std_precision_values = np.std(test_precision_values)

    print(f"Average test_clf_auc: {avg_test_precision}")
    print(f"Standard deviation of test_clf_auc: {std_precision_values}")


import pandas as pd
import os
import re

def process_csv_files(directory_path):
    # 初始化一个空的DataFrame来存储所有数据
    combined_df = pd.DataFrame()

    # 遍历指定目录下的所有文件
    for file_name in os.listdir(directory_path):
        if file_name.endswith(".csv"):
            # 读取CSV文件
            file_path = os.path.join(directory_path, file_name)
            df = pd.read_csv(file_path, index_col=0)  # 假设第一列是索引
            
            # 从行索引中提取方法名并作为新的一列
            df['Method'] = [re.match(r"(.*?)_\d+", idx).group(1) for idx in df.index]
            
            # 将当前DataFrame添加到总的DataFrame中
            combined_df = pd.concat([combined_df, df])

    # 计算每个方法的test_clf_auc的平均值和标准差
    results = combined_df.groupby('Method')['test_precision@20'].agg(['mean', 'std']).reset_index()
    results.columns = ['Method', 'Avg_test_precision', 'Std_test_precision']  # 重命名列以更清晰地表示它们的含义

    return results

# 指定目录路径
import sys
backbone = sys.argv[1]
print(sys.argv[1])
directory_path = f'/usr/scratch/jzhu617/renaissance/result/{backbone}_plbind/all_inherent'
# 调用函数并打印结果
results_df = process_csv_files(directory_path)
print(results_df)




import torch
import torch.nn as nn
from torch_scatter import scatter
import numpy as np
from torch_sparse import transpose
from torch_geometric.data import Data, Batch
from torch_geometric.utils import degree
from ..base import BaseRandom


def relabel(x, edge_index, batch, pos=None):

    num_nodes = x.size(0)
    sub_nodes = torch.unique(edge_index)
    x = x[sub_nodes]
    batch = batch[sub_nodes]
    row, col = edge_index
    # remapping the nodes in the explanatory subgraph to new ids.
    node_idx = row.new_full((num_nodes,), -1)
    node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=x.device)
    edge_index = node_idx[edge_index]
    if pos is not None:
        pos = pos[sub_nodes]
    return x, edge_index, batch, pos


def split_batch(g, edge_index):
    split = degree(g.batch[edge_index[0]], dtype=torch.long).tolist()
    edge_indices = torch.split(edge_index, split, dim=1)
    num_nodes = degree(g.batch, dtype=torch.long)
    cum_nodes = torch.cat([g.batch.new_zeros(1), num_nodes.cumsum(dim=0)[:-1]])
    num_edges = torch.tensor([e.size(1) for e in edge_indices], dtype=torch.long).to(g.x.device)
    cum_edges = torch.cat([g.batch.new_zeros(1), num_edges.cumsum(dim=0)[:-1]])

    return edge_indices, num_nodes, cum_nodes, num_edges, cum_edges


class DIR(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'dir'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device
        self.epoch = None
        self.causal_loss_coef = config['causal_loss_coef']
        self.conf_loss_coef = config['conf_loss_coef']
        self.reg_loss_coef = config['reg_loss_coef']
        self.causal_ratio = config['causal_ratio']
        self.split_data = 'old'
        self.test_flag = True

    def __loss__(self, causal_pred, conf_pred, clf_labels, epoch):
        detach_conf_out = conf_pred.clone().detach()

        causal_loss = self.criterion(causal_pred, clf_labels.float())
        conf_loss = self.criterion(conf_pred, clf_labels.float())

        env_loss = torch.tensor([]).to(conf_pred.device)
        for single_conf_pred in detach_conf_out:
            combine_pred = torch.sigmoid(single_conf_pred) * causal_pred
            tmp = self.criterion(combine_pred, clf_labels.float())
            env_loss = torch.cat([env_loss, tmp.unsqueeze(0)])
        reg_loss = env_loss.mean() + torch.var(env_loss * causal_pred.size(0))
        # all_loss = [self.criterion(conf_pred[i], clf_labels[i].float()) for i in range(conf_pred.shape[0])]

        causal_loss = self.causal_loss_coef * causal_loss
        conf_loss = self.conf_loss_coef * conf_loss
        reg_loss = self.reg_loss_coef * (epoch ** 1.6) * reg_loss

        loss = causal_loss + conf_loss + reg_loss
        loss_dict = {'loss': loss.item(), 'pred': causal_loss.item(), 'conf': conf_loss.item(), 'reg': reg_loss.item()}
        return loss, loss_dict

    def forward_pass(self, data, epoch, do_sampling, **kwargs):
        emb, edge_index = self.clf.get_emb(data)
        # original_clf_logits = self.clf(data, edge_attr=data.edge_attr)
        node_weight = self.extractor(emb, batch=data.batch, pool_out_lig=None)
        pred_edge_weight = self.node_attn_to_edge_attn(node_weight, edge_index)

        if self.split_data == 'continuous':
            pred_edge_weight = pred_edge_weight.view(-1)
            device = data.x.device
            causal_edge_weight = pred_edge_weight.clone()
            spu_edge_weight = -pred_edge_weight.clone()

            edge_indices, _, _, num_edges, cum_edges = split_batch(data)
            for edge_index, N, C in zip(edge_indices, num_edges, cum_edges):
                n_reserve = int(self.causal_ratio * N)
                # single_mask = pred_edge_weight[C:C + N]
                single_mask_detach = pred_edge_weight[C:C + N].detach().cpu().numpy()
                # rank = np.argpartition(-single_mask_detach, n_reserve)
                # idx_reserve, idx_drop = rank[:n_reserve], rank[n_reserve:]

                idx_reserve = single_mask_detach.argsort()[-n_reserve:]
                idx_drop = single_mask_detach.argsort()[:-n_reserve]

                if self.test_flag:
                    causal_edge_weight[C:C + N][idx_reserve] = 1
                causal_edge_weight[C:C + N][idx_drop] = 0
                    # = torch.cat([causal_edge_weight, single_mask[idx_reserve]])
                spu_edge_weight[C:C + N][idx_reserve] = 0

                # causal_edge_attr = torch.cat([causal_edge_attr, edge_attr[idx_reserve]])
                # spu_edge_attr = torch.cat([spu_edge_attr, edge_attr[idx_drop]])

        else:
            assert self.split_data == 'old'
            pred_edge_weight = pred_edge_weight.view(-1)
            device = data.x.device
            causal_edge_index = torch.LongTensor([[], []]).to(device)
            causal_edge_weight = torch.tensor([]).to(device)
            # causal_edge_attr = torch.tensor([]).to(device)
            spu_edge_index = torch.LongTensor([[], []]).to(device)
            spu_edge_weight = torch.tensor([]).to(device)
            # spu_edge_attr = torch.tensor([]).to(device)

            edge_indices, _, _, num_edges, cum_edges = split_batch(data, edge_index)
            for edge_index, N, C in zip(edge_indices, num_edges, cum_edges):
                n_reserve = int(self.causal_ratio * N)

                # edge_attr = data.edge_attr[C:C + N]
                single_mask = pred_edge_weight[C:C + N]
                single_mask_detach = pred_edge_weight[C:C + N].detach()
                rank = np.argpartition(-single_mask_detach.cpu().numpy(), n_reserve)
                idx_reserve, idx_drop = rank[:n_reserve], rank[n_reserve:]
                # if debug:
                #     print(n_reserve)
                #     print(idx_reserve)
                #     print(idx_drop)
                causal_edge_index = torch.cat([causal_edge_index, edge_index[:, idx_reserve]], dim=1)
                spu_edge_index = torch.cat([spu_edge_index, edge_index[:, idx_drop]], dim=1)

                causal_edge_weight = torch.cat([causal_edge_weight, single_mask[idx_reserve]])
                spu_edge_weight = torch.cat([spu_edge_weight, -1 * single_mask[idx_drop]])

                # causal_edge_attr = torch.cat([causal_edge_attr, edge_attr[idx_reserve]])
                # spu_edge_attr = torch.cat([spu_edge_attr, edge_attr[idx_drop]])

            causal_x, causal_edge_index, causal_batch, causal_pos = relabel(data.x, causal_edge_index, data.batch, data.pos)
            spu_x, spu_edge_index, spu_batch, spu_pos = relabel(data.x, spu_edge_index, data.batch, data.pos)

            causal_data = Batch(batch=causal_batch, edge_index=causal_edge_index,
                                x=causal_x, pos=causal_pos)
            conf_data = Batch(batch=spu_batch, edge_index=spu_edge_index,
                                x=spu_x, pos=spu_pos)

        if self.split_data == 'continuous':
            causal_node_emb = self.clf.get_emb(data, edge_attn=causal_edge_weight.reshape(-1, 1))[0]
            conf_node_emb = self.clf.get_emb(data, edge_attn=spu_edge_weight.reshape(-1, 1))[0].detach()
            causal_out = self.clf.get_pred_from_emb(causal_node_emb, data.batch)
            conf_out = self.clf.get_pred_from_spu_emb(conf_node_emb, data.batch)
        else:
            causal_node_emb = self.clf.get_emb(causal_data, edge_attn=causal_data.edge_weight)[0]
            conf_node_emb = self.clf.get_emb(conf_data, edge_attn=conf_data.edge_weight)[0].detach()
            # causal_graph_emb = self.clf.pool(causal_node_emb, causal_data.batch)
            # conf_graph_emb = self.clf.pool(conf_node_emb, conf_data.batch)
            causal_out = self.clf.get_pred_from_emb(causal_node_emb, causal_data.batch)
            conf_out = self.clf.get_pred_from_spu_emb(conf_node_emb, conf_data.batch)

        loss, loss_dict = self.__loss__(causal_out, conf_out, data.y, epoch)
        combine_out = torch.sigmoid(conf_out) * causal_out
        # res_attn = pred_edge_weight
        res_attn = pred_edge_weight if hasattr(data, 'edge_label') else node_weight

        return loss, loss_dict, combine_out, res_attn.reshape(-1)

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        # edge_attn = 1 - (1 - src_attn) * (1 - dst_attn)
        edge_attn = src_attn * dst_attn
        # edge_attn = (src_attn + dst_attn) / 2
        return edge_attn



import torch
import torch.nn as nn
from torch_geometric.nn import BatchNorm


class MLP(nn.Sequential):
    def __init__(self, channels, dropout_p, norm_type, act_type):
        norm = self.get_norm(norm_type)
        act = self.get_act(act_type)

        m = []
        for i in range(1, len(channels)):
            m.append(nn.Linear(channels[i - 1], channels[i]))

            if i < len(channels) - 1:
                m.append(norm(channels[i]))
                m.append(act())
                m.append(nn.Dropout(dropout_p))

        super(MLP, self).__init__(*m)

    @staticmethod
    def get_norm(norm_type):
        if isinstance(norm_type, str) and 'batch' in norm_type:
            return BatchNorm
        elif norm_type == 'none' or norm_type is None:
            return nn.Identity
        else:
            raise ValueError('Invalid normalization type: {}'.format(norm_type))

    @staticmethod
    def get_act(act_type):
        if act_type == 'relu':
            return nn.ReLU
        elif act_type == 'silu':
            return nn.SiLU
        elif act_type == 'none':
            return nn.Identity
        else:
            raise ValueError('Invalid activation type: {}'.format(act_type))


class CoorsNorm(nn.Module):
    def __init__(self, eps = 1e-6, scale_init = 1.):
        super().__init__()
        self.eps = eps
        scale = torch.zeros(1).fill_(scale_init)
        self.scale = nn.Parameter(scale)

    def forward(self, coors):
        norm = coors.norm(dim = -1, keepdim = True)
        normed_coors = coors / norm.clamp(min = self.eps)
        return normed_coors #* self.scale


class ExtractorMLP(nn.Module):

    def __init__(self, hidden_size, config, use_lig_info, out_dim=None):
        super().__init__()
        self.use_lig_info = use_lig_info

        dropout_p = config['dropout_p']
        norm_type = config['norm_type']
        act_type = config['act_type']
        covar_dim = config.get('covar_dim', None)
        if covar_dim is None:
            out_dim = 1
        attn_dim = covar_dim**2+2 if out_dim is None else out_dim
        self.feature_extractor = MLP([hidden_size * 1, hidden_size * 2, hidden_size, attn_dim], dropout_p, norm_type, act_type)

    def forward(self, emb, batch, pool_out_lig=None):
        # node_emb = emb + pool_out_lig[batch] if self.use_lig_info else emb
        attn_log_logits = self.feature_extractor(emb)
        return attn_log_logits


class FeatEncoder(torch.nn.Module):

    def __init__(self, hidden_size, categorical_feat, scalar_feat, n_categorical_feat_to_use=-1, n_scalar_feat_to_use=-1):
        super().__init__()
        self.embedding_list = torch.nn.ModuleList()

        self.num_categorical_feat = len(categorical_feat)
        self.n_categorical_feat_to_use = self.num_categorical_feat if n_categorical_feat_to_use == -1 else n_categorical_feat_to_use
        self.num_scalar_feat_to_use = scalar_feat if n_scalar_feat_to_use == -1 else n_scalar_feat_to_use

        for i in range(self.n_categorical_feat_to_use):
            num_categories = categorical_feat[i]
            emb = torch.nn.Embedding(num_categories, hidden_size)
            self.embedding_list.append(emb)

        if self.num_scalar_feat_to_use > 0:
            assert n_scalar_feat_to_use == -1
            self.linear = torch.nn.Linear(self.num_scalar_feat_to_use, hidden_size)

        total_cate_dim = self.n_categorical_feat_to_use*hidden_size
        self.dim_mapping = torch.nn.Linear(total_cate_dim + hidden_size, hidden_size) if self.num_scalar_feat_to_use > 0 else torch.nn.Linear(total_cate_dim, hidden_size)

    def forward(self, x):
        x_embedding = []
        for i in range(self.n_categorical_feat_to_use):
            x_embedding.append(self.embedding_list[i](x[:, i].long()))

        if self.num_scalar_feat_to_use > 0:
            x_embedding.append(self.linear(x[:, self.num_categorical_feat:]))

        x_embedding = self.dim_mapping(torch.cat(x_embedding, dim=-1))
        return x_embedding

# def load_optimizer()

def get_optimizer(clf, extractor, optimizer_config, method_name, warmup):
    # pred_lr = method_config['pred_lr']
    # pred_wd = method_config['pred_wd']

    wp_lr = optimizer_config['clf_lr']
    wp_wd = optimizer_config['clf_wd']
    attn_lr = optimizer_config['exp_lr']
    attn_wd = optimizer_config['exp_wd']
    emb_lr = optimizer_config['clf_lr']
    emb_wd = optimizer_config['clf_wd']

    algo = torch.optim.Adam
    # clf_emb_model_params = [kv[1]  for kv in clf.named_parameters() if 'emb_model' in kv[0]]
    clf_base_params = [kv[1] for kv in clf.named_parameters() if 'emb_model' not in kv[0]]
    clf_emb_model_params = [kv[1]  for kv in clf.named_parameters() if 'emb_model' in kv[0]]

    if warmup:
        return algo([{'params': clf_base_params}], lr=wp_lr, weight_decay=wp_wd)
    if method_name == 'pgexplainer':
        return algo([{'params': extractor.parameters()}], lr=attn_lr, weight_decay=attn_wd)
    else: # for inherent models

        return algo([{'params': extractor.parameters(), 'lr': attn_lr, 'weight_decay': attn_wd},
                     {'params': clf_base_params, 'lr': emb_lr, 'weight_decay': emb_wd}, 
                     {'params': clf_emb_model_params, 'lr': emb_lr, 'weight_decay': emb_wd}])



import yaml
import json
import argparse
from tqdm import tqdm
from pathlib import Path
import torch
import torch.nn as nn
from torch.nn import functional as F
from eval import FidelEvaluation, AUCEvaluation, AucFidelity
from get_model import Model
from baselines import *
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, load_checkpoint, ExtractorMLP, get_optimizer
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
from statistics import mean
import numpy as np
import pandas as pd
import warnings
from datetime import datetime
warnings.filterwarnings("ignore")


def eval_one_batch(baseline, data, epoch):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradx', 'inter_grad', 'gnnexplainer']):
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP
        loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=False)
        # from sklearn.metrics import roc_auc_score
        # roc_auc_score(infer_clf_logits, data.y.cpu())
        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)


def eval_one_epoch(baseline, data_loader, epoch, phase, seed, signal_class, writer=None, metric_list=None):
    pbar, avg_loss_dict = tqdm(data_loader), dict()
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    for idx, data in enumerate(pbar):
        # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
        loss_dict, clf_logits, attn = eval_one_batch(baseline, data.to(baseline.device), epoch)
        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()
        # print()
        eval_dict = {metric.name: metric.collect_batch(ex_labels, attn, data, signal_class, 'geometric')
                     for metric in metric_list}
        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
        fidel_list = [eval_dict[k] for k in eval_dict if 'fid' in k]
        eval_dict.update({'mean_fid': mean(fidel_list) if fidel_list else 0})

        desc = log_epoch(seed, epoch, phase, loss_dict, eval_dict, phase_info='dataset')
        pbar.set_description(desc)
        # compute the avg_loss for the epoch desc
        exec('for k, v in loss_dict.items():\n\tavg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)')

    epoch_dict = {eval_metric.name: eval_metric.eval_epoch() for eval_metric in metric_list} if metric_list else {}
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    epoch_fidel = [epoch_dict[k] for k in epoch_dict if 'fid' in k]
    epoch_dict.update({'mean_fid': mean(epoch_fidel) if epoch_fidel else 0})

    # log_epoch(seed, epoch, phase, avg_loss_dict, epoch_dict, writer)

    return epoch_dict


def test(config, method_name, exp_method, model_name, backbone_seed, method_seed, dataset_name, model_dir, device, metric_str):
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    writer = None
    log_dir = model_dir.parent() / "post-inherent"
    log_dir.mkdir(parents=True, exist_ok=True)
    print('The logging directory is', log_dir), print('=' * 80)

    batch_size = config['optimizer']['batch_size']
    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0, num_workers=8)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],  # backbone_config
                method_name, config[method_name],  # method_config
                dataset).to(device)

    extractor = ExtractorMLP(config[model_name]['hidden_size'], config[method_name], config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits

    # establish the model and the metrics
    backbone = eval(name_mapping[method_name])(clf, extractor, criterion, config[method_name]) \
        if method_name in inherent_models else clf
    assert load_checkpoint(backbone, model_dir, model_name=method_name, seed=backbone_seed, map_location=torch.device('cpu') if not torch.cuda.is_available() else None)

    if 'label_perturb' in exp_method:
        model = LabelPerturb(backbone, mode=int(exp_method[-1]))
    elif 'self_exp' == exp_method:
        model = backbone
    elif exp_method == 'pgexplainer':
        model = PGExplainer(backbone, extractor, criterion, config['pgexplainer'])
    else:
        model = eval(name_mapping[exp_method])(clf, criterion, config[exp_method])


    # metric_list = [AUCEvaluation()] + [FidelEvaluation(backbone, i/10) for i in range(2, 9)] + \
    #  [FidelEvaluation(backbone, i/10, instance='pos') for i in range(2, 9)] + \
    metric_list = []
    for one_metric in metric_str:
        if 'exp_auc' == one_metric:
            metric_list += [AUCEvaluation()]
        elif 'pos_fid' == one_metric:
            metric_list += [FidelEvaluation(backbone, i / 10, instance='pos', symbol='+') for i in reversed(range(2, 9))]
        elif 'neg_fid' == one_metric:
            metric_list += [FidelEvaluation(backbone, i / 10, instance='pos', symbol='-') for i in reversed(range(2, 9))]
        else:
            raise ValueError(f"Metric name {one_metric} besides exp_auc/pos_fid/neg_fid is not expected")

    # print('Use random explanation and fidelity w/ signal nodes to test the Model Sensitivity.')
    set_seed(method_seed, method_name=method_name, backbone_name=model_name)
    # metric_names = [i.name for i in metric_list] + ['clf_acc', 'clf_auc']
    # train_dict = eval_one_epoch(model, loaders['train'], 1, 'train', backbone_seed, signal_class, writer, metric_list)
    # valid_dict = eval_one_epoch(model, loaders['valid'], 1, 'valid', backbone_seed, signal_class, writer, metric_list)
    if exp_method == 'pgexplainer' and method_name != 'erm':
        optimizer = get_optimizer(clf, extractor, config['optimizer'], method_name, warmup=False)
        raise NotImplementedError('PGExplainer for inherent models are not implemented yet.')
    #
    else:
        test_dict = eval_one_epoch(model, loaders['test'], 1, 'test', backbone_seed, signal_class,  writer, metric_list)

    return {}, {}, test_dict

def save_multi_bseeds(multi_methods_res, method_name, exp_method, seeds, metric_str, exp_metric_str):
    # save_seeds = seeds +
    indexes = ['_'.join([method_name, str(seed), item]) for seed in seeds  for item in ['avg', 'std']]
    # from itertools import product
    # indexes = ['_'.join(item) for item in product(methods, seeds)]
    df = pd.DataFrame(multi_methods_res, index=indexes)

    day_dir = Path('../result') / config_name / 'sensitivity' / metric_str / datetime.now().strftime("%m_%d")
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([exp_method, method_name]) + '.csv')
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    return df

def get_avg_std_report(reports):
    # print(reports)
    all_keys = {k: [] for k in reports[0]}
    for report in reports:
        for k in report:
            all_keys[k].append(report[k])
    avg_report = {k: np.mean(v) for k, v in all_keys.items()}
    std_report = {k: np.std(v) for k, v in all_keys.items()}
    avg_std_report = {k: f'{np.mean(v):.3f} +/- {np.std(v):.3f}' for k, v in all_keys.items()}
    return avg_report, std_report, avg_std_report

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack')
    parser.add_argument('--clf_method', type=str, nargs="+", help='classification method used', default=['erm'])
    parser.add_argument('--exp_method', type=str, nargs="+", help='explanation method used', default=['label_perturb1', 'label_perturb0'])
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--bseeds', type=int, nargs="+", help='random seed for training backbone', default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    parser.add_argument('--mseeds', type=int, nargs="+", help='random seed for explainer', default=list(range(10)))
    parser.add_argument('--metrics', type=str, nargs="+", default=['exp_auc', 'pos_fid', 'neg_fid'])

    args = parser.parse_args()
    # sub_metric = 'avg_loss'
    print(args)

    dataset_name, method_name, model_name, cuda_id, note = args.dataset, args.clf_method, args.backbone, args.cuda, args.note

    device = torch.device(f'cuda:{cuda_id}' if cuda_id >= 0 and torch.cuda.is_available() else 'cpu')

    exp_method = args.exp_method
    config_name = '_'.join([model_name, dataset_name])
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('../configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    if config[method_name].get(model_name, False):
        config[method_name].update(config[method_name][model_name])

    print('=' * 80)
    print(f'Config for {method_name}: ', json.dumps(config[method_name], indent=4))
    # the directory is like egnn_actstrack / bseed0 / lri_bern

    model_dir = Path('../log') / config_name / method_name
    multi_bseeds_res = []
    for backbone_seed in args.bseeds:
        mseeds_res = []
        # cuda_id = backbone_seed % 5
        for method_seed in args.mseeds:
            # method_seed = int(method_seed)
            # print('int or str:', type(method_seed))
            train_report, valid_report, test_report = test(config, method_name, exp_method, model_name, backbone_seed, method_seed, dataset_name, model_dir, device, metric_str=args.metrics)
            mseeds_res += [test_report]
            # print('Train Dataset Result: ', json.dumps(train_report, indent=4))
            # print('Valid Dataset Result: ', json.dumps(valid_report, indent=4))
            # print('Test Dataset Result: ', json.dumps(test_report, indent=4))
        avg_report, std_report, avg_std_report = get_avg_std_report(mseeds_res)
        multi_bseeds_res += [avg_report, std_report]
        print(json.dumps(avg_std_report, indent=4))
    save_multi_bseeds(multi_bseeds_res, method_name, exp_method, args.bseeds, args.metric, args.exp_metric)




import sys
sys.path.append('../')

import os
import yaml
import shutil
import os.path as osp
from tqdm import tqdm

import numpy as np
import pandas as pd
import torch
from torch_geometric.data import InMemoryDataset, Data

import rdkit.Chem as Chem
import rdkit.Chem.AllChem as AllChem

from ..utils import download_url, extract_zip, decide_download


class SynMol(InMemoryDataset):
    ATOM_TYPES = ['C', 'N', 'O', 'S', 'F', 'P', 'Cl', 'Br', 'Na', 'Ca', 'I', 'B', 'H', '*']

    def __init__(self, root, data_config, transform, device='cpu', seed=0):
        self.url_raw = 'https://zenodo.org/record/7265547/files/synmol_raw.zip'
        self.url_processed = 'https://zenodo.org/record/7265547/files/synmol_processed.zip'
        self.seed = seed

        super().__init__(root, transform=transform)
        self.data, self.slices, self.idx_split = torch.load(self.processed_paths[0])
        self.data = self.data.to(device)
        self.x_dim = self.data.x.shape[1]
        self.pos_dim = self.data.pos.shape[1]
        self.feature_type = data_config['feature_type']
        self.signal_class = 1
        self.dataset_name = 'synmol'

        node_categorical_feat = [len(self.ATOM_TYPES)]
        if self.feature_type == 'only_pos':
            node_scalar_feat = self.pos_dim
            node_categorical_feat = []
        elif self.feature_type == 'only_x':
            node_scalar_feat = self.x_dim - 1
        elif self.feature_type == 'only_ones':
            node_scalar_feat = 1
            node_categorical_feat = []
        else:
            assert self.feature_type == 'both_x_pos'
            node_scalar_feat = self.x_dim - 1 + self.pos_dim

        self.feat_info = {'node_categorical_feat': node_categorical_feat, 'node_scalar_feat': node_scalar_feat}

    @property
    def raw_file_names(self):
        return [f'logic8_traintest_indices.npz', f'logic8_smiles.csv', 'true_raw_attribution_datadicts.npz', 'x_true.npz', 'y_true.npz']

    @property
    def processed_file_names(self):
        return ['data.pt']

    def download(self):
        if osp.exists(self.processed_paths[0]):
            return
        if decide_download(self.url_raw, is_raw=True):
            path = download_url(self.url_raw, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
        else:
            if decide_download(self.url_processed, is_raw=False):
                path = download_url(self.url_processed, self.root)
                extract_zip(path, self.root)
                os.unlink(path)
            else:
                print('Stop downloading.')
                shutil.rmtree(self.root)
                exit(-1)

    def process(self):
        if decide_download(self.url_processed, is_raw=False):
            path = download_url(self.url_processed, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
            return

        all_y = np.load(self.raw_dir + '/y_true.npz', allow_pickle=True)['y']
        all_x = np.load(self.raw_dir + '/x_true.npz', allow_pickle=True)['datadict_list'][0]
        all_exp_labels = np.load(self.raw_dir + '/true_raw_attribution_datadicts.npz', allow_pickle=True)['datadict_list']
        mol_df = pd.read_csv(self.raw_dir + '/logic8_smiles.csv')

        raw_idx_split = dict(np.load(self.raw_dir + '/logic8_traintest_indices.npz', allow_pickle=True))
        assert len(all_y) == len(all_x) == len(all_exp_labels) == (raw_idx_split['train_index'].shape[0] + raw_idx_split['test_index'].shape[0])
        split_dict = self.get_split_dict(raw_idx_split, seed=self.seed)

        data_list = []
        idx_split = {'train': [], 'valid': [], 'test': []}
        cnt = 0
        for idx, data in tqdm(enumerate(all_x), total=len(all_x)):
            x = torch.tensor(data['nodes'])
            x = torch.where(x == 1)[1].reshape(-1, 1)
            y = torch.tensor(all_y[idx]).reshape(-1, 1)

            node_label = torch.tensor(all_exp_labels[idx][0]['nodes'][:, -1]).view(-1)
            if all_exp_labels[idx][0]['nodes'].shape[1] > 1:
                assert np.all((all_exp_labels[idx][0]['nodes'][:, :-1].sum(axis=1) > 0) == (all_exp_labels[idx][0]['nodes'][:, -1] == 1))

            mol = Chem.MolFromSmiles(mol_df.iloc[idx]['smiles'])
            m = Chem.AddHs(mol)
            message_id = AllChem.EmbedMolecule(m, randomSeed=self.seed)
            if message_id < 0:
                print(f'Failed to embed molecule {idx}')
                continue
            message_id = AllChem.MMFFOptimizeMolecule(m, maxIters=1000)
            if message_id < 0:
                print(f'Failed to optimize molecule {idx}')
                continue
            m = Chem.RemoveHs(m)
            pos = torch.tensor(m.GetConformer().GetPositions(), dtype=torch.float)
            assert x.shape[0] == m.GetNumAtoms()
            for j in range(m.GetNumAtoms()):
                assert self.ATOM_TYPES[x[j]] == m.GetAtomWithIdx(j).GetSymbol() or m.GetAtomWithIdx(j).GetSymbol() not in self.ATOM_TYPES

            data_list.append(Data(x=x, pos=pos, y=y, node_label=node_label, mol_df_idx=idx))
            idx_split[split_dict[idx]].append(cnt)
            cnt += 1

        data, slices = self.collate(data_list)
        torch.save((data, slices, idx_split), self.processed_paths[0])

    @staticmethod
    def get_split_dict(raw_idx_split, seed):
        np.random.seed(seed)

        train_val_idx = raw_idx_split['train_index']
        idx = np.arange(len(train_val_idx))
        np.random.shuffle(idx)

        train_idx = train_val_idx[idx[:-1000]]
        valid_idx = train_val_idx[idx[-1000:]]
        test_idx = raw_idx_split['test_index']

        split_dict = {}
        for idx in train_idx:
            split_dict[idx] = 'train'
        for idx in valid_idx:
            split_dict[idx] = 'valid'
        for idx in test_idx:
            split_dict[idx] = 'test'
        return split_dict


if __name__ == '__main__':
    data_config = yaml.safe_load(open('../configs/egnn_synmol.yml'))['data']
    dataset = SynMol(root='../../data/synmol', data_config=data_config, seed=42)



import json
import argparse
import os.path
from datetime import datetime
import numpy as np
import torch
from trainer import run_one_seed
import pandas as pd
import warnings
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution
from pathlib import Path
warnings.filterwarnings("ignore")


def get_avg_std_report(reports):
    # print(reports)
    all_keys = {k: [] for k in reports[0]}
    for report in reports:
        for k in report:
            all_keys[k].append(report[k])
    avg_report = {k: np.mean(v) for k, v in all_keys.items()}
    std_report = {k: np.std(v) for k, v in all_keys.items()}
    avg_std_report = {k: f'{np.mean(v):.3f} \pm {np.std(v):.3f}' for k, v in all_keys.items()}
    return avg_report, std_report, avg_std_report


def save_multi_result(method_res, method_name, all_methods, bseed, seeds, config_name):
    seeds = [str(i) for i in seeds]
    aux_names = seeds + ['avg', 'std'] if len(seeds) > 1 else seeds
    indexes = [method_name+'_'+str(name) for name in aux_names]

    df = pd.DataFrame(method_res, index=indexes)

    day_dir = Path('result') / config_name / '_'.join(all_methods)
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join(['bs'+str(bseed)] + all_methods + seeds) + '.csv')
    with open(csv_dir, mode='a') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    now_df = pd.read_csv(csv_dir, index_col=0)
    return now_df


def main(args):
    initial_method_str = args.methods
    if len(args.methods) == 1:
        if args.methods[0] == 'all_inherent':
            args.methods = inherent_models
        elif args.methods[0] == 'all_attribution':
            args.methods = post_hoc_attribution
        elif args.methods[0] == 'all_explainer':
            args.methods = post_hoc_explainers # 'pgmexplainer',  post_hoc_explainers
        elif args.methods[0] == 'explainer_inherent':
            args.methods = ['gnnexplainer_lri_gaussian', 'gnnexplainer_lri_bern', 'pgexplainer_lri_gaussian', 'pgexplainer_lri_bern']
        else:
            print(args.methods)
            if args.methods[0] in inherent_models + post_hoc_attribution + post_hoc_explainers:
                pass
            else:
                raise ValueError(f'Unknown experiment {args.methods[0]} except for explainer/attribution/inherent.')

    #* set method seeds
    if args.seeds == -1:
        if set(args.methods) <= set(inherent_models):
            args.seeds = [args.bseed]
        elif set(args.methods) <= set(post_hoc_attribution):
            args.seeds = [0]
        elif set(args.methods) <= set(post_hoc_explainers):
            args.seeds = list(range(0, 10))
        elif any('pgexplainer_' in element or 'gnnexplainer_' in element for element in args.methods):
            args.seeds = [args.bseed]
        else:
            raise ValueError(f"Don't try to running methods of different kinds together. {set(args.methods)}")
    else:
        print(f'You have set seed {args.seeds} for {args.methods}.')

    if args.gpu_ratio is not None:
        torch.cuda.set_per_process_memory_fraction(args.gpu_ratio, args.cuda)
    multi_methods_res = []
    config_name = '_'.join([args.backbone, args.dataset])

    #* method preparation
    for method in args.methods:
        args.method = method
        multi_seeds_res, multi_seeds_attn = [], None
        save_dir = Path('log') / config_name / method / f"bs{args.bseed}_ms{args.seeds}_attns.csv"

        for seed in args.seeds:
            args.seed = seed
            report_dict, attn_df = run_one_seed(args, None)
            if multi_seeds_attn is None:
                multi_seeds_attn = attn_df
            else:
                # four columns are: node_labels, graph_labels, batch_idx, graph_idx
                assert multi_seeds_attn.iloc[:, -4:].equals(attn_df.iloc[:, -4:])
                multi_seeds_attn.insert(seed-args.seeds[0], seed, attn_df[seed])
            multi_seeds_res += [report_dict]
            # print(json.dumps(report_dict, indent=4))

        if not os.path.exists(save_dir) or args.rewrite:
            multi_seeds_attn.to_csv(save_dir)
        elif args.methods == inherent_models or args.methods == post_hoc_attribution or args.methods == post_hoc_explainers:
            print('='*80, f'Notice: You have rewrite the attns.csv for all_inherent/attribution/explainer.')
        else:
            raise FileExistsError(f'{save_dir} exists and you do not set rewrite.')

        avg_report, std_report, avg_std_report = get_avg_std_report(multi_seeds_res)
        multi_seeds_res += [avg_report, std_report] if len(args.seeds) > 1 else []
        print(f'[{len(args.seeds)} seeds for {method} on {args.dataset} (classifier training seed: {args.bseed}) done]\n',
              json.dumps(avg_std_report, indent=4))
        multi_methods_df = save_multi_result(multi_seeds_res, method, initial_method_str, args.bseed, args.seeds, config_name)

    print("=" * 80), print(multi_methods_df)


if __name__ == '__main__':
    time = datetime.now().strftime("%m_%d-%H_%M")
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used')
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--method', type=str, help='method used')
    parser.add_argument('--seed', type=int, help='random seed')
    parser.add_argument('--methods', type=str, help='methods tested', nargs='+')
    parser.add_argument('--seeds', type=int, nargs='+', help='random seed', default=-1)

    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--gpu_ratio', type=float, help='gpu memory ratio', default=None)
    parser.add_argument('--bseed', type=int, help='random seed for training backbone', default=0)
    parser.add_argument('--quick', action="store_true", help='ignore some evaluation')
    parser.add_argument('--save', action="store_true", help='save all erm models')
    parser.add_argument('--no_tqdm', action="store_true", help='disable the tqdm')
    parser.add_argument('--num_workers', type=int, default=0)
    parser.add_argument('--rewrite', action="store_true", help='rewrite the attns')
    parser.add_argument('--only_clf', action="store_true", help='train erm models and then exit')


    args = parser.parse_args()
    use_tqdm = False if args.no_tqdm else True
    # main_metric = 'exp_auc'
    # sub_metric = 'avg_loss'
    main(args)



import torch
import torch.nn as nn
from torch_geometric.data import Data, Batch
from torch_geometric.utils import degree

def relabel(x, edge_index, batch, pos=None):

    num_nodes = x.size(0)
    sub_nodes = torch.unique(edge_index)
    x = x[sub_nodes]
    batch = batch[sub_nodes]
    row, col = edge_index
    # remapping the nodes in the explanatory subgraph to new ids.
    node_idx = row.new_full((num_nodes,), -1)
    node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=x.device)
    edge_index = node_idx[edge_index]
    if pos is not None:
        pos = pos[sub_nodes]
    return x, edge_index, batch, pos

def split_batch(g):
    split = degree(g.batch[g.edge_index[0]], dtype=torch.long).tolist()
    edge_indices = torch.split(g.edge_index, split, dim=1)
    num_nodes = degree(g.batch, dtype=torch.long)
    cum_nodes = torch.cat([g.batch.new_zeros(1), num_nodes.cumsum(dim=0)[:-1]])
    num_edges = torch.tensor([e.size(1) for e in edge_indices], dtype=torch.long).to(g.x.device)
    cum_edges = torch.cat([g.batch.new_zeros(1), num_edges.cumsum(dim=0)[:-1]])

    return edge_indices, num_nodes, cum_nodes, num_edges, cum_edges

class Test(nn.Module):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'test'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device
        self.epoch = 0
        self.causal_loss_coef = 1
        self.conf_loss_coef = 0
        self.reg_loss_coef = 0
        self.causal_ratio = 1

    def __loss__(self, causal_pred, conf_pred, clf_labels):
        causal_loss = self.criterion(causal_pred, clf_labels.float())

        loss = causal_loss
        loss_dict = {'loss': loss.item(), 'causal': causal_loss.item()}
        return loss, loss_dict


    def warming(self, data):
        clf_logits = self.clf(data)
        pred_loss = self.criterion(clf_logits, data.y.float())
        return pred_loss, clf_logits

    def select_epoch(self):
        imp = 1
        return imp

    def explaining(self, data, x_level, **kwargs):

        emb = self.clf.get_emb(data, edge_attr=data.edge_attr)
        detach_emb = emb.clone().detach()
        org_logits = self.clf.get_pred_from_emb(emb, data.batch)
        # original_clf_logits = self.clf(data, edge_attr=data.edge_attr)
        attach_node_weight = self.extractor(emb, data, att_type='node')
        detach_node_weight = self.extractor(detach_emb, data, att_type='node')
        attach_edge_weight = self.node_attn_to_edge_attn(attach_node_weight, data.edge_index)
        detach_edge_weight = self.node_attn_to_edge_attn(detach_node_weight, data.edge_index)
        # pred_edge_weight[:] = 1
        # pred_edge_weight = self.linear(edge_rep).view(-1)
        causal_edge_weight = detach_edge_weight.clone()
        conf_edge_weight = detach_edge_weight.clone()

        causal_edge_weight[:] = 1 if self.epoch % 2 == 1 else -1
        # causal_edge_weight[:] = self.select_epoch()
        conf_edge_weight[:] = 1
        for i in range(4):
            causal_out = self.clf(data, edge_attn=causal_edge_weight.reshape(-1, 1))
        # causal_out = self.clf.get_pred_from_emb(emb, data.batch)
        # causal_out = self.clf(data, edge_attn=causal_edge_weight.reshape(-1, 1))
        conf_out = self.clf(data, edge_attn=conf_edge_weight.reshape(-1, 1))

        loss, loss_dict = self.__loss__(causal_out, 0, data.y)

        # res_attn = pred_edge_weight
        res_attn = attach_edge_weight if hasattr(data, 'edge_label') else attach_node_weight

        return loss, loss_dict, org_logits, res_attn.reshape(-1)


    def create_new_data(self, graph, idx, edge_weight):
        x = graph.x
        edge_index = graph.edge_index[:, idx]
        edge_attr = graph.edge_attr[idx] if graph.edge_attr is not None else None

        # node relabel
        num_nodes = x.size(0)
        sub_nodes = torch.unique(edge_index)

        x = x[sub_nodes]

        pos = graph.pos[sub_nodes] if graph.pos is not None else None
        # pos = pos

        row, col = edge_index
        # remapping the nodes in the explanatory subgraph to new ids.
        node_idx = row.new_full((num_nodes,), -1)
        node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=row.device)
        edge_index = node_idx[edge_index]

        return Data(x=x, edge_index=edge_index, pos=pos, edge_attr=edge_attr, edge_weight=edge_weight)

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        # edge_attn = 1 - (1 - src_attn) * (1 - dst_attn)
        edge_attn = src_attn * dst_attn
        # edge_attn = (src_attn + dst_attn) / 2
        return edge_attn



import os
pre_cuda_id = int(os.environ.get('CUDA_VISIBLE_DEVICES')) if os.environ.get('CUDA_VISIBLE_DEVICES') else None
# os.environ['CUDA_VISIBLE_DEVICES'] = '0,1,2,3,4,5'
import yaml
import json
import pandas as pd
import shutil
import argparse
from tqdm import tqdm
from pathlib import Path
from copy import deepcopy
from itertools import product
# import nni
import numpy as np
import torch
import torch.nn as nn
from torch.nn import functional as F
# from torch.utils.tensorboard import SummaryWriter
from .eval import FidelEvaluation, AUCEvaluation, PrecEvaluation
# from .get_model import Model
from .baselines import *
from .get_data import get_data_loaders
from .utils import to_cpu, log_epoch, set_seed, update_and_save_best_epoch_res, load_checkpoint, save_checkpoint, ExtractorMLP, get_optimizer
from .utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
from statistics import mean
import warnings
warnings.filterwarnings("ignore")


def select_augmentation(dataset):
    if dataset == 'plbind':
        def func(data, phase, data_loader, batch_idx, **kwargs):
            # only used for PLBind dataset
            loader_len = len(data_loader)
            neg_aug_p = 0.1
            # neg_aug_p = data_config['neg_aug_p']
            if neg_aug_p and np.random.rand() < neg_aug_p and batch_idx != loader_len - 1 and phase in ['train', 'warm']:
                aug_data = next(iter(data_loader))
                data.x_lig = aug_data.x_lig
                data.pos_lig = aug_data.pos_lig
                data.x_lig_batch = aug_data.x_lig_batch
                data.y = torch.zeros_like(data.y)
            return data
    else:
        def func(data, phase, data_loader, batch_idx, **kwargs):
            return data
    return func


def eval_one_batch(baseline, optimizer, data, epoch, phase):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradx', 'inter_grad', 'gnnexplainer']):
        assert optimizer is None
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # BernMaskP
        do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP
        if phase == 'warm':
            loss, loss_dict, infer_clf_logits, node_attn = baseline.warming(data)
        else:
            loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=do_sampling)

        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)


def train_one_batch(baseline, optimizer, data, epoch, phase):
    baseline.extractor.train() if hasattr(baseline, 'extractor') else None
    baseline.clf.train() if (baseline.name != 'pgexplainer' or phase == 'warm') else baseline.clf.eval()

    if phase == 'warm':
        loss, loss_dict, org_clf_logits, node_attn = baseline.warming(data)
    else:
        loss, loss_dict, org_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=True)
    
    # if baseline.name == 'test_inherent':
    #     loss.backward(retain_graph=True)
    #     data.node_grads += data.pos.grad.norm(dim=1, p=2)
    #     data.edge_grads += data.edge_attn.grad
    # else:
    optimizer.zero_grad()

    torch.set_printoptions(precision=8)
    loss.backward()

    optimizer.step()

    return loss_dict, to_cpu(org_clf_logits), to_cpu(node_attn)


def run_one_epoch(baseline, optimizer, data_loader, epoch, phase, seed, signal_class, writer=None, metric_list=None, return_attn=False, data_augmentation=None):
    loader_len = len(data_loader)
    use_tqdm = True
    run_one_batch = train_one_batch if optimizer else eval_one_batch
    pbar, avg_loss_dict = tqdm(data_loader) if use_tqdm else data_loader, dict()
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    save_epoch_attn = []
    for idx, data in enumerate(pbar):
        # torch.cuda.empty_cache()
        if data_augmentation is not None:
            data = data_augmentation(data, phase, data_loader, idx, loader_len=loader_len)
        loss_dict, clf_logits, attn = run_one_batch(baseline, optimizer, data, epoch, phase)
        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()

        # prepare to save attn
        if return_attn:
            graph_labels = data.y[data.batch]
            batch_idx = torch.full_like(graph_labels, idx)
            graph_idx = data.batch.unsqueeze(-1)
            save_attn = torch.cat([attn.unsqueeze(-1), ex_labels.unsqueeze(-1), graph_labels, batch_idx, graph_idx], dim=1)
            save_epoch_attn.append(save_attn)

        eval_dict = {metric.name: metric.collect_batch(ex_labels, attn, data, signal_class, 'geometric')
                     for metric in metric_list} if metric_list else {}
        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
        batch_fid = [eval_dict[k] for k in eval_dict if 'fid' in k and 'all' in k]
        eval_dict.update({'mean_fid': mean(batch_fid)}) if phase in ['valid', 'test'] and batch_fid else {}

        desc = log_epoch(seed, epoch, phase, loss_dict, eval_dict)
        pbar.set_description(desc) if use_tqdm else None
        # compute the avg_loss for the epoch desc
        for k, v in loss_dict.items():
            avg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)

    epoch_dict = {eval_metric.name: eval_metric.eval_epoch() for eval_metric in metric_list} if metric_list else {}
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    fid_score = [epoch_dict[k] for k in epoch_dict if 'fid' in k and 'all' in k]
    epoch_dict.update({'mean_fid': mean(fid_score)}) if phase in ['valid', 'test'] and fid_score else {}

    log_epoch(seed, epoch, phase, avg_loss_dict, epoch_dict, writer)

    if return_attn:
        save_epoch_attn = torch.cat(save_epoch_attn, dim=0)
        return epoch_dict, save_epoch_attn
    else:
        return epoch_dict


def train(config, method_name, model_name, backbone_seed, seed, dataset_name, parent_dir, device, main_metric, args, data_augmentation):
    # writer = SummaryWriter(log_dir) if log_dir is not None else None
    quick, save, num_workers = args.quick, args.save, args.num_workers
    writer = None
    model_dir, log_dir = (parent_dir / method_name, ) * 2 if method_name in inherent_models \
        else (parent_dir / 'erm', None) if model_name in post_hoc_attribution \
        else (parent_dir / 'erm', parent_dir / method_name)
    #! specially for post-hoc explaining inherent models
    log_dir = (parent_dir / method_name) if method_name.startswith(('gnnexplainer_', 'pgexplainer_')) else log_dir
    # log_dir = parent_dir / method_name if method_name in inherent_models else None
    log_dir.mkdir(parents=True, exist_ok=True) if log_dir is not None else None
    model_dir.mkdir(parents=True, exist_ok=True)

    batch_size = config['optimizer']['batch_size'] if (method_name, dataset_name) != ('gnnlrp', 'plbind') else 1
    model_cofig = config[method_name] if method_name in inherent_models else config['erm']
    warmup = model_cofig['warmup']
    epochs = config[method_name]['epochs'] if not method_name.startswith(('gnnexplainer_', 'pgexplainer_')) else config[method_name.split('_', 1)[0]]['epochs'] 
    data_config = config['data']
    # notice: in get_data_loaders some seed may be defined and changed
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0, num_workers=num_workers, device=device)
    signal_class = dataset.signal_class
    backbone_config = config[model_name] 

    #! specially for post-hoc explaining inherent models
    if not method_name.startswith(('pgexplainer_', 'gnnexplainer_')):
        method_config = config[method_name]

        clf = Model(model_name, backbone_config, method_name, method_config, dataset).to(device)

        extractor = ASAPooling(backbone_config['hidden_size'], ratio=method_config['casual_ratio'], dropout=method_config['dropout_p']) \
            if method_name == 'asap' else \
            ExtractorMLP(backbone_config['hidden_size'], method_config, config['data'].get('use_lig_info', False)) \
            if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
        extractor = extractor.to(device)
        constructor = eval(name_mapping[method_name])
        optimizer = get_optimizer(clf, extractor, config['optimizer'], method_name, warmup=True)

        if args.save:
            optimizer = get_optimizer(clf, extractor, config['optimizer'], method_name, warmup=True)

    criterion = F.binary_cross_entropy_with_logits
    map_location = torch.device('cpu') if not torch.cuda.is_available() else device
    init_metric_list = [AUCEvaluation()] if dataset_name != 'plbind' else [PrecEvaluation(20)]

    # establish the model and the metrics
    if method_name in inherent_models:
        baseline = constructor(clf, extractor, criterion, config[method_name])
        for epoch in range(1, warmup+1):
            run_one_epoch(baseline, optimizer, loaders['train'], epoch, 'warm', seed, signal_class, writer, data_augmentation=data_augmentation)
            run_one_epoch(baseline, None, loaders['valid'], epoch, 'warm', seed, signal_class, writer, data_augmentation=data_augmentation)
            run_one_epoch(baseline, None, loaders['test'], epoch, 'warm', seed, signal_class, writer, data_augmentation=data_augmentation)

        metric_list = init_metric_list
    elif method_name in post_hoc_attribution + post_hoc_explainers:
        baseline = constructor(clf, criterion, config[method_name]) if method_name != 'pgexplainer' else PGExplainer(clf, extractor, criterion, config['pgexplainer'])
        if args.save or not load_checkpoint(baseline.clf, model_dir, model_name='erm', seed=backbone_seed, map_location=map_location):
            best_auc, best_acc = 0, 0
            for epoch in range(1, warmup + 1):
                run_one_epoch(baseline, optimizer, loaders['train'], epoch, 'warm', backbone_seed, signal_class, writer, data_augmentation=data_augmentation)
                epoch_dict = run_one_epoch(baseline, None, loaders['valid'], epoch, 'warm', backbone_seed, signal_class, writer)
                clf_acc, clf_auc = epoch_dict['clf_acc'], epoch_dict['clf_auc']
                better_valid = (clf_auc > best_auc) or ((clf_auc == best_auc) and clf_acc > best_acc)
                if better_valid:
                    best_auc, best_acc = clf_auc, clf_acc
                    #* Save_info is not None means that need to save all intermidiate erm models
                    clf_acc, clf_auc = epoch_dict['clf_acc'], epoch_dict['clf_auc']
                    save_info = {'epoch': epoch, 'acc': clf_acc, 'auc': clf_auc} if save else None
                    save_checkpoint(baseline.clf, model_dir, model_name='erm', backbone_seed=backbone_seed,
                                    seed=backbone_seed, save_info=save_info)
            # save_checkpoint(baseline.clf, model_dir, model_name='erm', backbone_seed=backbone_seed, seed=backbone_seed)
        metric_list =  init_metric_list + [FidelEvaluation(baseline.clf, i/10) for i in range(2, 9)] + \
                  [FidelEvaluation(baseline.clf, i/10, instance='pos') for i in range(2, 9)] + \
                  [FidelEvaluation(baseline.clf, i/10, instance='neg') for i in range(2, 9)] if quick == False else \
                  [AUCEvaluation()] + [FidelEvaluation(baseline.clf, i/10) for i in range(2, 9)]
        baseline.start_tracking() if 'grad' in method_name or method_name == 'gnnlrp' else None
    else: # for post-hoc explaining inherent models.
        exp_method, clf_method = method_name.split('_', 1)
        #* ReConfig
        inher_method_cfg = config[clf_method]
        post_method_cfg = config[exp_method]

        inher_extractor = ASAPooling(backbone_config['hidden_size'], ratio=inher_method_cfg['casual_ratio'], dropout=inher_method_cfg['dropout_p']) if clf_method == 'asap' else ExtractorMLP(backbone_config['hidden_size'], inher_method_cfg, config['data'].get('use_lig_info', False))

        clf = Model(model_name, backbone_config, clf_method, inher_method_cfg, dataset).to(device)
        clf = eval(name_mapping[clf_method])(clf, inher_extractor, criterion, inher_method_cfg)

        #* load inherent model...
        assert load_checkpoint(clf, parent_dir / clf_method , model_name=clf_method, seed=backbone_seed, map_location=map_location)

        if method_name.startswith("pg"):

            extractor = ExtractorMLP(backbone_config['hidden_size'], post_method_cfg, config['data'].get('use_lig_info', False))

            #! change names for the following training
            method_name = 'pgexplainer'
            baseline = PGExplainer(clf, extractor, criterion, config['pgexplainer']).to(device)
            # optimizer = get_optimizer(clf_model, pg_extractor, config['optimizer'], "pgexplainer", warmup=False)
        else:
            assert method_name.startswith('gnne'), f"Unknown Method {method_name}"
            extractor = nn.Identity() 

            #! change name for training
            method_name = 'gnnexplainer'
            baseline = GNNExplainer(clf, criterion, config['gnnexplainer']).to(device)

        metric_list = init_metric_list + [FidelEvaluation(baseline.clf, i/10) for i in range(2, 9)] + \
            [FidelEvaluation(baseline.clf, i/10, instance='pos') for i in range(2, 9)] + \
            [FidelEvaluation(baseline.clf, i/10, instance='neg') for i in range(2, 9)] if quick == False else \
            [AUCEvaluation()] + [FidelEvaluation(baseline.clf, i/10) for i in range(2, 9)]
#! --------- End of this branch -----------------------


    if args.save or args.only_clf:
        if method_name not in inherent_models:
            print("Intermidiate ERM models have been saved.")
            exit(0)
        else:
            pass


    two_encoder = not config[method_name].get('one_encoder', True)
    if two_encoder:
        clf.emb_model = deepcopy(clf.model)

    if method_name not in inherent_models:
        set_seed(seed, method_name=method_name, backbone_name=model_name)

    metric_names = [a + b for a, b in product(['valid_', 'test_'], [i.name for i in metric_list]+['clf_acc', 'clf_auc', 'mean_fid'])]
    # metric_names = [j+i.name for i in metric_list for j in ['valid_', 'test_']]
    metric_dict = {}.fromkeys(metric_names, -1)
    metric_dict['best_epoch'] = 1
    best_attn = None
    optimizer = get_optimizer(clf, extractor, config['optimizer'], method_name, warmup=False)
    no_better_val = False
    early_stop_count = 0
    # torch.cuda.empty_cache()
    for epoch in range(1, epochs+1):
        torch.cuda.empty_cache()
        if args.save:
            #! print("This script is for saving   checkpoints for each epoch for inherent methods")        
            #* weak_trn for the second round training to complement those inherent models...
            weak_training = True
            if method_name in ['vgib', 'asap']:
                weak_training = 'minor'
            optimizer = get_optimizer(clf, extractor, config['optimizer'], method_name, warmup=False, weak_training=weak_training)
           #* train and valid one epoch
            run_one_epoch(baseline, optimizer, loaders['train'], epoch, 'train', seed, signal_class, writer, data_augmentation=data_augmentation) 
            test_dict = run_one_epoch(baseline, None, loaders['test'], epoch, 'test', seed, signal_class, writer, metric_list)
            #* This is used for args.save...
            clf_acc, clf_auc, exp_auc = test_dict['clf_acc'], test_dict['clf_auc'], test_dict['exp_auc']
            save_info = {'epoch': epoch, 'acc': clf_acc, 'auc': clf_auc, 'xauc': exp_auc}
            save_checkpoint(baseline, model_dir, model_name=method_name, backbone_seed=backbone_seed,
                                    seed=backbone_seed, save_info=save_info)
        # no args.save means the normal logic
        elif method_name in inherent_models + ['pgexplainer']:
            #* train one epoch
            run_one_epoch(baseline, optimizer, loaders['train'], epoch, 'train', seed, signal_class, writer, data_augmentation=data_augmentation)
            #* valid one epoch
            valid_dict = run_one_epoch(baseline, None, loaders['valid'], epoch, 'valid', seed, signal_class, writer, metric_list)
            test_dict, epoch_attn = run_one_epoch(baseline, None, loaders['test'], epoch, 'test', seed, signal_class, writer, metric_list, return_attn=True)
        else:
            test_dict, epoch_attn = run_one_epoch(baseline, None, loaders['test'], epoch, 'test', seed, signal_class,  writer, metric_list, return_attn=True)
            valid_dict = test_dict # other methods don't need validation to select epochs
        
        if args.save:
            if epoch == epochs:
                print("Complete saving all checkpoints for inherent methods")
                exit(0)
            else:
                continue

        # print(metric_dict)
        metric_dict, new_best = update_and_save_best_epoch_res(baseline, metric_dict, valid_dict, test_dict, epoch, log_dir, backbone_seed, seed, writer, method_name, main_metric)
        best_attn = epoch_attn if new_best else best_attn
        metric_dict.update({'default': metric_dict[f'valid_{main_metric}']})
        nni.report_intermediate_result(metric_dict)

    meta_index = 'attn' if method_name in post_hoc_attribution + inherent_models else seed
    indexes = [meta_index, 'node_labels', 'graph_labels', 'batch_idx', 'graph_idx']

    return metric_dict, (best_attn, indexes)


def run_one_seed(args, optimized_params):
    print(args)

    dataset_name, method_name, model_name, cuda_id, note = args.dataset, args.method, args.backbone, args.cuda, args.note
    method_seed, backbone_seed = args.seed, args.bseed
    main_metric = 'clf_auc' if method_name in inherent_models + ['test'] else 'mean_fid'

    config_name = '_'.join([model_name, dataset_name])
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('./configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    # if config[method_name].get('pred_lr', False):
        # config['optimizer'].update({'pred_lr': config[method_name]['pred_lr']})
    if optimized_params:
        config[method_name].update(optimized_params)
    data_augmentation = select_augmentation(dataset_name)
    print('=' * 80)
    print(f'Config for {method_name}: ', json.dumps(config.get(method_name, f"{{{method_name}: None}}"), indent=4))
    # :{pre_cuda_id}
    # if not os.environ.get('CUDA_VISIBLE_DEVICES'):
    #     os.environ['CUDA_VISIBLE_DEVICES'] = f'{cuda_id}'
    cuda_ = f'cuda' if cuda_id == 99 else f'cuda:{cuda_id}' if cuda_id >= 0 else 'cpu'
    device = torch.device(cuda_)
    # log_dir = None
    # if config['logging']['tensorboard'] or method_name in ['gradcam', 'gradgeo', 'bernmask_p', 'bernmask']:
    main_dir = Path('log') / config_name
    main_dir.mkdir(parents=True, exist_ok=True)
        # shutil.copy(config_path, log_dir / config_path.name)
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    report_dict, (best_attn, indexes) = train(config, method_name, model_name, backbone_seed, method_seed, dataset_name, main_dir, device, main_metric, args=args, data_augmentation=data_augmentation)
    attn_df = pd.DataFrame(best_attn, columns=indexes)

    return report_dict, attn_df


def main(args):
    if args.gpu_ratio is not None:
        torch.cuda.set_per_process_memory_fraction(args.gpu_ratio, args.cuda)

    params = nni.get_next_parameter()
    # if args.multi_seeds_tune:
    #     report_dict = {}
    #     for seed in [0, 1, 2]:
    #         args.bseed = seed
    #         args.seed = seed
    #         one_report, _ = run_one_seed(args, params)
    #         report_dict.update({k: v+report_dict.get(k, 0) for k, v in one_report.items()})
    # else:
    report_dict, _ = run_one_seed(args, params)

    print(json.dumps(report_dict, indent=4))
    # nni_report = dict(report_dict, **{'default': report_dict[f'valid_{main_metric}']})
    nni.report_final_result(report_dict)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack')
    parser.add_argument('-m', '--method', type=str, help='method used', default='lri_gaussian')
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--seed', type=int, help='random seed', default=0)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--gpu_ratio', type=float, help='gpu memory ratio', default=None)
    parser.add_argument('--bseed', type=int, help='random seed for training backbone', default=0)
    parser.add_argument('--quick', action="store_true", help='ignore some evaluation')
    parser.add_argument('--save', action="store_true", help='save all erm models')
    parser.add_argument('--only_clf', action="store_true", help='train erm models and then exit')
    parser.add_argument('--no_tqdm', action="store_true", help='disable the tqdm')
    parser.add_argument('--num_workers', type=int, default=0)
    # parser.add_argument('--multi_seeds_tune', action="store_true")

    exp_args = parser.parse_args()
    use_tqdm = False if exp_args.no_tqdm else True
    # sub_metric = 'avg_loss'
    main(exp_args)



import os
import sys
import re
import random

path = sys.argv[-1]
# 获取该目录下所有文件，存入列表中
fileList = os.listdir(path)
# keys = list(range(725, 875))
name_dict = dict([(key, []) for key in range(650, 890)])

for filename in fileList:
    if filename == 'before':
        continue
    list_ = re.split(r'[\./-]', filename)
    auc_item = [i for i in list_ if 'auc' in i]
    auc = int(auc_item[0][:3])
    if auc in name_dict:
        name_dict[auc] += [filename]

for auc, value in name_dict.items():
    if len(value) == 0:
        continue
    else:
        number = random.randint(0, len(value)-1)
        for i, filename in enumerate(value):
            oldname = path + os.sep + filename
            if number == i:
                path_list = re.split(r'[\./-]', filename)
                path_list.pop(-1)
                path_list.insert(0, path_list.pop(2))
                new_filename = '-'.join(path_list) + '.pt'
                newname = path + os.sep + new_filename
                # newname = oldname
            else:
                newname = path + os.sep + 'no.pt'
            os.rename(oldname, newname)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scienceplots
import seaborn as sns
from scipy.stats import pearsonr, linregress
plt.style.use(['science', 'notebook'])
import os
model_var_path = r"C:\Users\86173\Desktop\NEWMAIN\model_var\\"
save_path = r"C:\Users\86173\Desktop\NEWMAIN"


def generate_reliability_df(name_list, path, backbone_name):
    df = pd.DataFrame(columns=['Model', 'Backbone', 'quantification', 'ken_corr', 'spr_corr', 'pr_corr', 'avg_auc'])
    init_df = pd.read_csv(f"{path}/label_perturb1.csv", index_col=0)
    init_df['ACCfidpos'] = (init_df['ACCfidpos'] + 1) / 2
    for row_name in init_df.index:
        com_df = pd.DataFrame(columns=['exp_auc', 'ACCfidpos'])
        for name in name_list:
            csv_name = path + '\\' + name + '.csv'
            method_df = pd.read_csv(csv_name, index_col=0)
            method_df['ACCfidpos'] = (method_df['ACCfidpos'] + 1) / 2

            com_df.loc[len(com_df)] = [method_df.loc[row_name, 'exp_auc'], method_df.loc[row_name, 'ACCfidpos']]
        ken, spr, pr = com_df.corr('kendall').iloc[1, 0], com_df.corr('spearman').iloc[1, 0], com_df.corr('pearson').iloc[1, 0]
        avg_exp_auc = com_df['exp_auc'].mean()
        quantification = init_df.loc[row_name, 'ACCfidpos']
        df.loc[len(df)] = [row_name, backbone_name, quantification, ken, spr, pr, avg_exp_auc]
    return df

def generate_df_from_method_list(name_list, column, path, dataset_name):
    df = pd.DataFrame(columns=['Method', 'variable', 'exp_auc'])
    if dataset_name == 'synmol':
        int_list = [750, 800, 850, 900, 950, 1000]
        variables = [0.775, 0.825, 0.875, 0.925, 0.975]
    elif dataset_name == 'actstrack':
        int_list = [500, 600, 700, 800, 900, 1000]
        variables = [0.550, 0.650, 0.750, 0.850, 0.950]
    elif dataset_name == 'tau3mu':
        int_list = [650, 690, 730, 770, 810, 850]
        variables = [0.670, 0.710, 0.750, 0.790, 0.830]
    # for name in name_list:
    for name in name_list:
        csv_name = path + '\\' + name + '.csv'
        method_df = pd.read_csv(csv_name, index_col=0)
        method_df = method_df[~method_df.index.duplicated()]
        data_series = method_df[column]
        count = 0
        for index_name in method_df.index:
            auc = int(index_name[:3])
            for i in range(1, len(int_list)):
                if int_list[i-1] < auc <= int_list[i]:
                    df.loc[len(df)] = [name, variables[i - 1], data_series[index_name]]
                    # new_row = pd.DataFrame({'Method': name, 'variable': variables[i - 1], 'exp_auc': data_series[index_name]}, index=[0])
                    # df = pd.concat([df, new_row], ignore_index=True)
    return df


# plot notions https://www.notion.so/Notions-for-plot-0793de0cd418443786c80e0fcca30ad1
def draw_boxplot(ax, backbone, dataset, method_list, column, data_path, xlabel='Classification AUC', ylabel='Explanation AUC', legend=True):
    # plot
    # if 'exp' in column:
    # elif 'fid' in column:
    #     node_type = column[-3:]
    #     ylabel = f'Fidelity on {node_type.capitalize()} Samples'

    data = generate_df_from_method_list(method_list, column, data_path, dataset_name=dataset)
    sns.boxplot(data=data, x='variable', y='exp_auc', hue='Method', ax=ax,
                        showfliers=True,
                        showmeans=False,
                        # meanprops={'markerfacecolor': 'C0', 'markeredgecolor': 'white', 'marker': 'o'},
                        medianprops={"color": "white", "linewidth": 0.5},
                        boxprops={"edgecolor": "white", "linewidth": 0.5},
                        whiskerprops={"linewidth": 1.5},
                        capprops={"linewidth": 1.5})
    # VP = ax.boxplot(data,
    #                 patch_artist=True,
    #                 showfliers=True,
    #                 showmeans=True,
    #                 meanprops={'markerfacecolor': 'C0', 'markeredgecolor': 'white', 'marker': 'o'},
    #                 medianprops={"color": "white", "linewidth": 0.5},
    #                 boxprops={"facecolor": "C0", "edgecolor": "white",
    #                           "linewidth": 0.5},
    #                 whiskerprops={"color": "C0", "linewidth": 1.5},
    #                 capprops={"color": "C0", "linewidth": 1.5})
    ax.set(xlabel=xlabel, ylabel=ylabel, title=backbone.upper())
    if legend is False:
        ax.legend([])


def draw_box_for_one_dataset(axes, method_list, backbones, dataset):
    dataset_path = model_var_path + dataset + '/'
    ylabels = ["Explanation AUC", '', '']
    xlabels = ['', 'Classification AUC', '']
    legends = [True, False, False]
    for i, backbone in enumerate(backbones):
        draw_boxplot(axes[i], backbone, dataset, method_list, 'exp_auc', xlabel=xlabels[i], ylabel=ylabels[i], legend=legends[i], data_path=dataset_path+backbone)
    return


def corr_sensitivity_fidelity(value, path='C:\\Users\86173\Desktop\\0710collect\synmol\egnn\post-hoc'):
    df = pd.DataFrame()
    for root, ds, fs in os.walk(path):
        for f in fs:
            seed = f[5] # like bseed0
            fullname = os.path.join(root, f)
            data = pd.read_csv(fullname, index_col=0)
            sub_indexes = [i for i in data.index if i[-1] == seed or 'grad' in i or 'lrp' in i or 'label' in i ]
            data = data.loc[sub_indexes]
            data.set_index('exp_method', inplace=True)

            # partial dataframe
            df[seed] = data[value]
    print(df.T.corr('pearson'))
    print(df.T.corr('kendall'))
    print(df.T.corr('spearman'))


def draw_reliability_interval_box(axs, dataset, ordered_backbones, base_path=r'C:\Users\86173\Desktop\NEWMAIN\model_rely', coef='pearson', y='reliability'):
    dataset_path = f'{base_path}/{dataset}'
    df_list = []
    for i, backbone in enumerate(ordered_backbones):
        one_path = f'{dataset_path}/{backbone}'
        methods_list = name_list_1 if y == 'auc' else name_list_1 + ['label_perturb1']
        one_df = generate_reliability_df(methods_list, one_path, backbone)
        df = one_df.set_index('Model')
        # try:
        df["q_range"] = pd.cut(df['quantification'], bins=3, precision=2)
        # except:
        print()
        df_list += [df]
    all_df = pd.concat(df_list)
    y_name = 'avg_auc' if y == 'auc' else 'ken_corr' if coef == 'kendall' else 'spr_corr' if coef == 'spearman' else 'pr_corr'
    y_label = 'Average Explanation ROC AUC' if y == 'auc' else f'{coef.capitalize()} Correlation'
    all_df.sort_values(by=['Backbone', 'q_range'], inplace=True, ascending=[False, True])
    sns.boxplot(data=all_df, x='q_range', y=y_name, hue='Backbone', ax=axs,
                showfliers=True,
                showmeans=False,
                # meanprops={'markerfacecolor': 'C0', 'markeredgecolor': 'white', 'marker': 'o'},
                medianprops={"color": "white", "linewidth": 0.5},
                boxprops={"edgecolor": "white", "linewidth": 0.5},
                whiskerprops={"linewidth": 1.5},
                capprops={"linewidth": 1.5})
    axs.tick_params(axis='x', labelrotation=30)
    axs.set(xlabel='Value Interval', ylabel=y_label, title=dataset.upper())


def draw_reliability_plots(axs, dataset, ordered_backbones, base_path=r'C:\Users\86173\Desktop\NEWMAIN\model_rely', coef='pearson', y='reliability'):
    dataset_path = f'{base_path}/{dataset}'
    df_list = []
    y_name = 'avg_auc' if y == 'auc' else 'ken_corr' if coef == 'kendall' else 'spr_corr' if coef == 'spearman' else 'pr_corr'
    y_label = 'Average Explanation ROC AUC' if y == 'auc' else f'{coef.capitalize()} Correlation'
    for i, backbone in enumerate(ordered_backbones):
        one_path = f'{dataset_path}/{backbone}'
        methods_list = name_list_1 if y == 'auc' else name_list_1+['label_perturb1']
        one_df = generate_reliability_df(methods_list, one_path, backbone)
        df = one_df.set_index('Model')

        axs.scatter(x=df['quantification'], y=df[y_name], label=backbone)
        # df["q_range"] = pd.cut(df['quantification'], bins=4, labels=[1, 2, 3, 4])
        df_list += [df]
    all_df = pd.concat(df_list)
    use_df = all_df[['quantification', y_name]]
    # text = f"Spearman coefficient: {use_df.corr('spearman').iloc[1, 0]:.2f}"
    # axs.text(0.05, 0.95, text, ha='left', va='top', transform=axs.transAxes)
    axs.legend()
        # sns.boxplot(data=all_df, x='q_range', y='pr_corr', hue='Backbone', ax=ax,
        #             showfliers=True,
        #             showmeans=False,
        #             # meanprops={'markerfacecolor': 'C0', 'markeredgecolor': 'white', 'marker': 'o'},
        #             medianprops={"color": "white", "linewidth": 0.5},
        #             boxprops={"edgecolor": "white", "linewidth": 0.5},
        #             whiskerprops={"linewidth": 1.5},
        #             capprops={"linewidth": 1.5})
    axs.set(xlabel='Label Fidelity AUC', ylabel=y_label, title=dataset.upper())


if __name__ == '__main__':
    backbones = ['egnn', 'dgcnn', 'pointtrans']
    name_list_1 = ['gradcam', 'inter_grad', 'gradx', 'gnnlrp']
    ordered_name_list = [name_list_1]
    datasets = ['synmol', 'actstrack', 'tau3mu']

    def explanation_prediction_corr():
        fig, axs = plt.subplots(nrows=len(datasets), ncols=3, sharex='row', sharey='row', figsize=(24, 6*len(datasets)))
        for index, dataset in enumerate(datasets):
            # corr_sensitivity_fidelity('ACCfid-pos')  # ['ACCfid+pos', 'ACCfid-pos', 'ACCfidpos']
            draw_box_for_one_dataset(axs[index], name_list_1, backbones, dataset)
        fig.tight_layout()
        fig.savefig(f'{save_path}/explanation_prediction_corr.pdf', dpi=300, transparent=True)
    # draw_boxplot(name, 'valid_fid_pos', post_fix='png')
    # explanation_prediction_corr()

    def reliability_value_quantification(datasets, type=('scatter', 'box'), y_axis='auc'):
        if 'scatter' in type:
            fig, axs = plt.subplots(nrows=1, ncols=len(datasets), figsize=(8*len(datasets), 6 * 1))
            for index, dataset in enumerate(datasets):
                draw_reliability_plots(axs[index], dataset, ordered_backbones=['pointtrans', 'egnn', 'dgcnn'], coef='pearson', y=y_axis)
            fig.tight_layout()
            fig.savefig(f'{save_path}/sca_reliability.pdf', dpi=300, transparent=True)
        if 'box' in type:
            fig, axs = plt.subplots(nrows=1, ncols=len(datasets), figsize=(8*len(datasets), 6 * 1))
            for index, dataset in enumerate(datasets):
                draw_reliability_interval_box(axs[index], dataset, ordered_backbones=['pointtrans', 'egnn', 'dgcnn'], coef='pearson', y=y_axis)
            fig.tight_layout()
            fig.savefig(f'{save_path}/box_reliability.pdf', dpi=300, transparent=True)

    reliability_value_quantification(['synmol', 'actstrack'])




import yaml
import json
import argparse
from tqdm import tqdm
from pathlib import Path
from sklearn.metrics import roc_auc_score
import torch
import torch.nn as nn
from torch.nn import functional as F
from eval import FidelEvaluation, AUCEvaluation
from get_model import Model
from baselines import *
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, load_checkpoint, ExtractorMLP, get_optimizer
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
import numpy as np
import pandas as pd
import warnings
from datetime import datetime
from sklearn.metrics import accuracy_score
warnings.filterwarnings("ignore")


def eval_one_batch(baseline, data, epoch):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradx', 'inter_grad', 'gnnexplainer']):
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP

        loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=False)

        if baseline.name == 'vgib':
            infer_clf_logits = baseline.forward(data)
        # from sklearn.metrics import roc_auc_score
        # roc_auc_score(infer_clf_logits, data.y.cpu())
        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)


def model_trustworthy(config, method_name, model_name, backbone_seed, dataset_name, log_dir, device):
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    print('The logging directory is', log_dir), print('=' * 80)

    # batch_size = config['optimizer']['batch_size']
    batch_size = 1
    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],  # backbone_config
                method_name, config[method_name],  # method_config
                dataset).to(device)

    extractor = ExtractorMLP(config[model_name]['hidden_size'], config[method_name], config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits

    map_location = torch.device('cpu') if not torch.cuda.is_available() else None
    baseline = eval(name_mapping[method_name])(clf, extractor, criterion, config[method_name]) if method_name in inherent_models + ['pgexplainer'] \
        else eval(name_mapping[method_name])(clf, criterion, config[method_name])

    if method_name in inherent_models:
        assert load_checkpoint(baseline, log_dir, model_name=method_name, seed=backbone_seed, map_location=map_location,
                               backbone_seed=backbone_seed)
    elif method_name == 'pgexplainer':
        assert load_checkpoint(baseline, log_dir, model_name=method_name, seed=backbone_seed, map_location=map_location,
                        backbone_seed=backbone_seed)
    else:
        print(f'The method {method_name} has no extra model.')

    if method_name not in inherent_models:
        erm_dir = log_dir.parent / 'erm'
        assert load_checkpoint(baseline.clf, erm_dir, model_name='erm', seed=backbone_seed, map_location=map_location)
    # establish the model and the metrics

    baseline.start_tracking() if 'grad' in method_name or method_name == 'gnnlrp' else None
    metric_list = [AUCEvaluation()] if batch_size > 64 else None

    pbar, avg_loss_dict = tqdm(loaders['test']), dict()
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    epoch_attn, epoch_label, epoch_sig, epoch_bkg = [], [], [], []
    # sp_list = []
    for idx, data in enumerate(pbar):
        if data.y.item() == 0:
            continue

        eval_dict = {}        # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
        loss_dict, clf_logits, attn = eval_one_batch(baseline, data.to(baseline.device), 1)
        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()
        # print()
        # min_max_sclarer
        attn = (attn - attn.min()) / (attn.max() - attn.min()) if method_name not in inherent_models else attn

        epoch_attn.append(attn), epoch_label.append(ex_labels)
        epoch_sig.append(attn[torch.where(data.node_label == 1)]), epoch_bkg.append(attn[torch.where(data.node_label == 0)])

        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
        eval_dict.update({f'exp_auc': roc_auc_score(ex_labels, attn)})

        # print('The spearman correlation between avg and std:', )
        desc = log_epoch(backbone_seed, 1, 'test', loss_dict, eval_dict, phase_info='dataset')
        pbar.set_description(desc)
        # compute the avg_loss for the epoch desc
        exec('for k, v in loss_dict.items():\n\tavg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)')

    # epoch_dict = {eval_metric.name: eval_metric.eval_epoch(return_att=True) for eval_metric in metric_list} if metric_list else {}
    # assert or  metric_list[0].name == 'exp_auc'
    epoch_attn, epoch_label, epoch_sig, epoch_bkg = torch.cat(epoch_attn), torch.cat(epoch_label), torch.cat(epoch_sig), torch.cat(epoch_bkg)
    epoch_dict = {'exp_auc': roc_auc_score(epoch_label, epoch_attn)}
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    print('-' * 80, '\n', 'epoch_dict:', epoch_dict, '\n', '-' * 80)

    return epoch_attn, epoch_label, epoch_sig, epoch_bkg

def save_multi_bseeds(multi_methods_res, method_name, exp_method, seeds):
    seeds += ['avg', 'std']
    indexes = [method_name+'_'+str(seed) for seed in seeds]
    # from itertools import product
    # indexes = ['_'.join(item) for item in product(methods, seeds)]
    df = pd.DataFrame(multi_methods_res, index=indexes)

    day_dir = Path('../result') / config_name / datetime.now().strftime("%m_%d")
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([exp_method, method_name, 'sensitivity.csv']))
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    return df


def save_multi_result(method_res, all_methods, config_name, metric_str, exp_metric_str):

    df = pd.DataFrame(method_res, index=all_methods)

    day_dir = Path('../result') / config_name / 'model_trustworthy' / metric_str
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([exp_metric_str] + all_methods) + '_trust.csv')
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    now_df = pd.read_csv(csv_dir, index_col=0)
    return now_df


def average_accuracy(label, attn):
    zero_one = np.linspace(0, 1, num=50)
    scores = np.array([])
    for treshold in zero_one:
        pred = (attn > treshold).float()
        scores = np.append(scores, accuracy_score(label, pred))
    return scores.mean()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack')
    parser.add_argument('--method', type=str, help='method used')
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--seeds', type=int, nargs="+", help='random seed for exp methods pipeline', default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    # parser.add_argument('--mseed', type=int, help='random seed for explainer', default=0)
    parser.add_argument('--bseeds', type=int, nargs="+", help='random seed to establish a series of classifiers', default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    parser.add_argument('--metric', type=str, choices=['acc_fid_pos', 'auc_fid_all'], default='acc_fid_pos')
    parser.add_argument('--exp_metric', type=str, choices=['roc_auc', 'avg_acc'], default='avg_acc')

    parser.add_argument('--use_csv', action="store_true")
    parser.add_argument('--skip_py', action="store_true")

    args = parser.parse_args()
    # sub_metric = 'avg_loss'
    print(args)
    dataset_name, model_name, cuda_id, note = args.dataset, args.backbone, args.cuda, args.note

    if not args.skip_py:
        import os
        os.system(f"python test_sensitivity.py -d {dataset_name} -b {model_name} --metric {args.metric} --cuda {cuda_id} --exp_method label_perturb1")
        os.system(f"python test_sensitivity.py -d {dataset_name} -b {model_name} --metric {args.metric} --cuda {cuda_id} --exp_method label_perturb0")
    # The result is stored in the result/egnn_synmol/sensitivity/label_perturb1_erm.csv
    config_name = '_'.join([model_name, dataset_name])
    path1 = Path(
        '../result') / config_name / 'sensitivity' / args.metric / datetime.now().strftime("%m_%d") / 'label_perturb1_erm.csv'
    path0 = Path(
        '../result') / config_name / 'sensitivity' / args.metric / datetime.now().strftime("%m_%d") / 'label_perturb0_erm.csv'
    perturb1_score = pd.read_csv(path1, index_col=0)
    perturb0_score = pd.read_csv(path0, index_col=0)
    rows = [i for i in perturb1_score.index if 'avg' in i]
    erm_score = (perturb1_score['mean_fid'] - perturb0_score['mean_fid'])[rows]
    erm_dict = {k.replace('erm', 'bseed').replace('_avg', ''): v for k, v in erm_score.to_dict().items()}
    print(erm_dict)

    device = torch.device(f'cuda:{cuda_id}' if cuda_id >= 0 and torch.cuda.is_available() else 'cpu')
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('../configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    all_methods = post_hoc_attribution + ['pgexplainer', 'pgmexplainer'] if args.method == 'all' else [args.method]
    method_res = []
    metric_func = average_accuracy if args.exp_metric == 'avg_acc' else roc_auc_score if args.exp_metric == 'roc_auc' else None
    for method_name in all_methods:
        method_dict = {}
        method_seeds = [0] if method_name in post_hoc_attribution else args.seeds
        if config[method_name].get(model_name, False):
            config[method_name].update(config[method_name][model_name])
        print('=' * 80)
        print(f'Config for {method_name}: ', json.dumps(config[method_name], indent=4))

        model_dir = Path('../log') / config_name / method_name
        multi_attn, multi_sig_attn, multi_bkg_attn = None, None, None

        for one_seed in args.bseeds:
            # we directly load attns from csv file
            all_seeds = [0] if method_name in post_hoc_attribution else args.seeds
            assert args.use_csv
            save_dir = model_dir / f"bs{one_seed}_ms{all_seeds}_attns.csv"
            seed_df = pd.read_csv(save_dir, index_col=0)
            seed_df = seed_df[seed_df['graph_labels'] == 1]
            epoch_label = torch.tensor(seed_df['node_labels'].values)
            if 'attn' in seed_df:  # this means the method has no need to vary seeds for fixed backbone_seed
                avg_attn = torch.tensor(seed_df['attn'].values).unsqueeze(1)
            else:  # for post-hoc explainers, there are many seed res and only use the same seed with backbone_seed
                avg_attn = torch.tensor(seed_df[[str(seed) for seed in all_seeds]].values).mean(1, keepdim=True)
            method_dict[f'bseed_{one_seed}'] = metric_func(epoch_label, avg_attn)
        method_res += [method_dict]
        print(json.dumps(method_dict, indent=4))

    df = save_multi_result([erm_dict] + method_res, ['erm_score'] + all_methods, config_name, args.metric, args.exp_metric)
    sp_corr = df.T.corr(method="spearman")
    pr_corr = df.T.corr(method="pearson")
    kd_corr = df.T.corr(method='kendall')
    # print(f"Models' Score Func: {args.metric} \n"
    print(f"Score Func: {args.exp_metric}")
    print('spearman: \n', sp_corr)
    print('pearson: \n', pr_corr)
    print('kendall: \n', kd_corr)
    path = Path('../result') / config_name / 'model_trustworthy' / f'{args.metric}-{args.exp_metric}.csv'
    sp_corr.to_csv(path)





import copy
from sklearn.metrics import roc_auc_score
import numpy as np
import torch
from torch_geometric.data import Data, Batch
from abc import ABC

def fidelity(intepretation, explainer, sparsity=0.2, symbol='+'):
    model = explainer.baseline.clf
    node_weights = control_sparsity(intepretation.node_imp, sparsity, symbol)
    device = intepretation.edge_index.device

    ## construct new subdata
    graph = intepretation
    idx = node_weights.reshape(-1).nonzero().reshape(-1)
    if not idx.numel():
        raise ValueError("Please Check the selection of sparsity")
    edge_list = []
    # print(node_weights)
    for edge_pair in graph.edge_index.T:
        edge_list += [edge_pair] if edge_pair[0] in idx and edge_pair[1] in idx else []
    if edge_list:
        edge_index = torch.vstack(edge_list).T
    else:
        edge_index = torch.tensor([], dtype=graph.edge_index.dtype, device=graph.edge_index.device).reshape(2, -1)

    if graph.edge_attr is not None:
        edge_attr_list = [graph.edge_attr[i] if (idx == edge_pair[0]).numel() and (idx == edge_pair[1]).numel()
                            else None for i, edge_pair in enumerate(graph.edge_index.T)]
        edge_attr = torch.vstack(edge_attr_list)
    else:
        edge_attr = None

    x = graph.x[idx]
    if graph.pos is not None:
        pos = graph.pos[idx]
    else:
        pos = None

    row = edge_index[0]
    node_idx = row.new_full((max(idx)+1,), -1)
    node_idx[idx] = torch.arange(idx.size(0), device=idx.device)
    edge_index = node_idx[edge_index]

    data = Data(x=x, y=graph.y, pos=pos, edge_index=edge_index, edge_attr=edge_attr, x_lig=graph.x_lig if hasattr(graph, "x_lig") else None, pos_lig=graph.pos_lig if hasattr(graph, "pos_lig") else None)

    ## compute two forward
    with torch.no_grad():
        origin_logits = model(intepretation)
        masked_logits = model(data.to(device))

    ## compute probability difference
    clf_labels = intepretation.y.clone()
    clf_labels[clf_labels == 0] = -1

    origin_pred = origin_logits.sigmoid()
    masked_pred = masked_logits.sigmoid()
    score = ((origin_pred - masked_pred) * clf_labels).item()
    return score





def x_rocauc(interpretation):
    interpretation = interpretation.cpu()
    label = interpretation.node_label
    preds = interpretation.node_imp
    try:
        score = roc_auc_score(label, preds)
    except:
        score = -1
        raise ValueError("Cannot compute roc_auc, please check whether the node labels has non-zero values.")
    return score


class BaseEvaluation(ABC):
    # def collect_batch(self, x_labels, weights, data, signal_class, x_level):
    #     pass

    def reset(self):
        pass

    def collect_batch(self, *args, **kwargs):
        pass

    def eval_epoch(self):
        pass


def partial_weights(node_weights, pos_neg):

    pos_weights = copy.deepcopy(node_weights)
    neg_weights = copy.deepcopy(1 - node_weights)

    pos_idx = torch.where(pos_weights == 1)[0]
    neg_idx = torch.where(neg_weights == 1)[0]

    pos_zero_num = int((1-pos_neg[0]) * pos_weights.sum().item())
    neg_zero_num = int((1-pos_neg[1]) * neg_weights.sum().item())
    np.random.seed(99)
    torch.cuda.manual_seed(99)
    torch.manual_seed(99)

    pos_idx = pos_idx[torch.randperm(pos_idx.size(0))[:pos_zero_num]]
    neg_idx = neg_idx[torch.randperm(neg_idx.size(0))[:neg_zero_num]]

    weights = torch.ones_like(node_weights)
    weights[pos_idx] = 0
    weights[neg_idx] = 0
    return weights


class FidelEvaluation(BaseEvaluation):
    def __init__(self, model, sparsity, type='acc', symbol='+', instance='all'):
        self.sparsity = sparsity
        self.perf = []
        self.valid = []
        self.test = []
        self.type = type
        self.symbol = symbol
        self.name = self.type.upper()+'fid'+symbol+instance+'@'+str(sparsity)
        self.instance = instance
        self.classifier = model
        self.device = next(model.parameters()).device

    def create_new_data(self, data, weights, weight_type='edge', signal_class=None, instance=None):
        sum_ = 0
        data_list = []
        count = 0
        pos_data_list = []
        for graph in data.to_data_list():
            if weight_type == 'node':
                node_weights = weights[sum_:sum_ + graph.num_nodes]
                sum_ += graph.num_nodes
                if instance == 'pos' and graph.y.item() != signal_class:
                    continue
                if instance == 'neg' and graph.y.item() == signal_class:
                    continue

                node_weights = control_sparsity(node_weights, self.sparsity, self.symbol)
                idx = node_weights.reshape(-1).nonzero().reshape(-1)
                if not idx.numel():
                    continue
                edge_list = []
                # print(node_weights)
                for edge_pair in graph.edge_index.T:
                    edge_list += [edge_pair] if edge_pair[0] in idx and edge_pair[1] in idx else []
                if edge_list:
                    edge_index = torch.vstack(edge_list).T
                else:
                    edge_index = torch.tensor([], dtype=graph.edge_index.dtype, device=graph.edge_index.device).reshape(2, -1)

                if graph.edge_attr is not None:
                    edge_attr_list = [graph.edge_attr[i] if (idx == edge_pair[0]).numel() and (idx == edge_pair[1]).numel()
                                      else None for i, edge_pair in enumerate(graph.edge_index.T)]
                    edge_attr = torch.vstack(edge_attr_list)
                else:
                    edge_attr = None

                x = graph.x[idx]
                if graph.pos is not None:
                    pos = graph.pos[idx]
                else:
                    pos = None

                row = edge_index[0]
                node_idx = row.new_full((max(idx)+1,), -1)
                node_idx[idx] = torch.arange(idx.size(0), device=idx.device)
                edge_index = node_idx[edge_index]
            else:
                graph_weights = weights[sum_:sum_ + graph.num_edges]
                sum_ += graph.num_edges
                if instance == 'pos' and graph.y.item() != signal_class:
                    continue
                if instance == 'neg' and graph.y.item() == signal_class:
                    continue

                graph_weights = control_sparsity(graph_weights, self.sparsity, self.symbol)
                idx = graph_weights.reshape(-1).nonzero().reshape(-1)
                assert idx.numel()
                x = graph.x
                edge_index = graph.edge_index[:, idx]
                edge_attr = graph.edge_attr[idx] if graph.edge_attr is not None else None

                # node relabel
                num_nodes = x.size(0)
                sub_nodes = torch.unique(edge_index)

                x = x[sub_nodes]

                if graph.pos is not None:
                    pos = graph.pos
                    pos = pos[sub_nodes]
                else:
                    pos = None


                row, col = edge_index
                # remapping the nodes in the explanatory subgraph to new ids.
                node_idx = row.new_full((num_nodes,), -1)
                node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=row.device)
                edge_index = node_idx[edge_index]
            if hasattr(graph, "x_lig"):
                # print(graph)
                data_list += [Data(x=x, y=graph.y, pos=pos, edge_index=edge_index, edge_attr=edge_attr,
                                   x_lig=graph.x_lig, pos_lig=graph.pos_lig)]
            else:
                data_list += [Data(x=x, y=graph.y, pos=pos, edge_index=edge_index, edge_attr=edge_attr)]
            # data_list += [Data(x=x, y=graph.y, pos=pos, edge_index=edge_index, edge_attr=edge_attr)]
            pos_data_list += [graph]
        if "-fidelity" in self.name and count:
            print(f'There is {count} graphs with no edges in this batch. (len:{len(data_list)})')
        follow_batch = ['x_lig'] if hasattr(graph, "x_lig") else None

        if data_list and pos_data_list:
            org_data = Batch.from_data_list(pos_data_list, follow_batch=follow_batch)
            new_data = Batch.from_data_list(data_list, follow_batch=follow_batch)
            return org_data, new_data
        else:
            return None, None

    def collect_batch(self, x_labels, weights, data, signal_class, x_level):
        # data.edge_index = self.classifier.get_emb(data)[1]
        # print(data.edge_index is None)
        weights = weights.reshape(-1, 1)
        # pos_data, pos_new_data = self.get_pos_instances(data, weights, weight_type=weight_type)
        if hasattr(data, "edge_label"):
            weight_type = 'edge'
        elif x_level == 'geometric':
            weight_type = 'node'
        else:
            assert x_level == 'graph'
            weights = node_attn_to_edge_attn(weights, data.edge_index)
            weight_type = 'edge'
        # weight_type = 'node' if weights.shape[0] != data.edge_index.shape[1] else 'edge'
        # print(data.x.device, weights.device)
        pos_data, pos_new_data = self.create_new_data(data, weights, weight_type=weight_type, signal_class=signal_class, instance=self.instance)

        if pos_new_data is None:
            return -1

        # if weights.shape[0] != data.edge_index.shape[1]:    # node_weights
        #     weights = node_attn_to_edge_attn(weights, data.edge_index)
        with torch.no_grad():
            origin_logits = self.classifier(pos_data.to(self.device))
            masked_logits = self.classifier(pos_new_data.to(self.device))

        clf_labels = pos_data.y.clone()

        if self.type == 'prob':
            clf_labels[clf_labels == 0] = -1

            origin_pred = origin_logits.sigmoid()
            masked_pred = masked_logits.sigmoid()
            scores = (origin_pred - masked_pred) * clf_labels

        else:
            assert self.type == 'acc'
            # assert self.type == 'acc'
            origin_pred = (origin_logits.sigmoid() > 0.5).float()
            masked_pred = (masked_logits.sigmoid() > 0.5).float()
            scores = (origin_pred == clf_labels).float() - (masked_pred == clf_labels).float()

        self.perf.append(scores.reshape(-1))
        return scores.reshape(-1).mean().item()

    def eval_epoch(self):
        # in the phas 'train', the train_res will be -1
        if not self.perf:
            return -1
        else:
            perf = torch.cat(self.perf).cpu()

        if self.type == 'auc':
            auc_score = roc_auc_score()
        return perf.mean().item()

    def reset(self):
        self.perf = []

    def update_epoch(self, valid_res, test_res):
        self.valid.append(valid_res)
        self.test.append(test_res) if not test_res else None

        return self.valid, self.test


class PrecEvaluation(BaseEvaluation):

    def __init__(self, k) -> None:
        self.prec = []
        self.valid = []
        self.test = []
        self.k = k
        self.name = f'precision@{k}'
        self.scale = 'instance'

    def reset(self):
        self.prec = []

    def collect_batch(self, x_labels, node_att, data, signal_class, x_level):
        if node_att.dim() == 2:
            node_att = node_att.reshape(-1)
        x_labels, node_att, graph_ids = get_signal_class(x_labels, node_att, data, signal_class) 
        batch_precision = []
        for id in graph_ids.unique():
            extract = lambda data: data[graph_ids == id]
            gx_labels, gnode_att = extract(x_labels), extract(node_att)
            #* desending ranking top k
            ids_of_topk = np.argsort(-gnode_att) [:self.k]
            topk_x_labels = gx_labels[ids_of_topk]
            precision = topk_x_labels.sum().item() / self.k
            batch_precision.append(precision)

        self.prec += batch_precision
        if batch_precision:
            return sum(batch_precision) / len(batch_precision)
        else:
            return -1
    
    def eval_epoch(self):
        if not self.prec:
            return -1
        else:
            prec = self.prec
            return sum(prec) / len(prec)


class AUCEvaluation(BaseEvaluation):
    """
    A class enabling the evaluation of the AUC metric on both graphs and nodes.

    :param ground_truth: ground truth labels.
    :param indices: Which indices to evaluate.

    :funcion get_score: obtain the roc auc score.
    """

    def __init__(self):
        self.att = []
        self.gnd = []
        self.valid = []
        self.test = []
        self.scale = 'dataset'
        self.name = 'exp_auc'

    def collect_batch(self, x_labels, node_att, data, signal_class, x_level):
        x_labels, node_att, _ = get_signal_class(x_labels, node_att, data, signal_class)
        self.att.append(node_att)
        self.gnd.append(x_labels)
        # print(torch.where(torch.isnan(a), torch.full_like(a, 0), a))
        try:
            score = roc_auc_score(x_labels.cpu(), node_att.cpu())
        except:
            print("cannot compute roc_auc")
            score = -1
        return score


    def eval_epoch(self, return_att=False):
        # in the phase 'train', the train_res will be -1
        if not self.att:
            return -1
        else:
            att = torch.cat(self.att)
            label = torch.cat(self.gnd)
            return roc_auc_score(label.cpu(), att.cpu()) if not return_att else (att.cpu(), label.cpu())

    def reset(self):
        self.att = []
        self.gnd = []

    def update_epoch(self, valid_res, test_res):
        self.valid.append(valid_res)
        self.test.append(test_res) if not test_res else None
        return self.valid, self.test


def get_signal_class(x_labels, att, data, signal_class: int):
    # regard edges as a node set
    graph_id_of_node = data.batch.cpu()
    graph_id_of_edge = graph_id_of_node[data.edge_index[0]] if hasattr(data, 'edge_label') else None
    if len(x_labels) == data.num_nodes:
        graph_id = graph_id_of_node
    elif len(x_labels) == data.num_edges:
        graph_id = graph_id_of_edge

    graph_label = data.y.cpu()
    in_signal_class = (graph_label[graph_id] == signal_class).reshape(-1)

    return x_labels[in_signal_class], att[in_signal_class], graph_id[in_signal_class]


def control_sparsity(mask, sparsity=None, symbol='+'):
        r"""

        :param mask: mask that need to transform
        :param sparsity: sparsity_list we need to control i.e. 0.7, 0.5
        :return: transformed mask where top 1 - sparsity_list values are set to inf.
        """
        if sparsity is None:
            sparsity = 0.7

        # if len(mask.shape)>1:
        #     mask = mask.squeeze()
        _, indices = torch.sort(mask, dim=0, descending=True)
        mask_len = mask.shape[0]
        split_point = int((1 - sparsity) * mask_len)
        important_indices = indices[: split_point]
        unimportant_indices = indices[split_point:]
        trans_mask = mask.clone()
        if symbol == "+":  # larger indicates batter
            trans_mask[important_indices] = 0
            trans_mask[unimportant_indices] = 1
        else:
            assert symbol == '-' #lower indicates better
            trans_mask[important_indices] = 1
            trans_mask[unimportant_indices] = 0


        return trans_mask


def node_attn_to_edge_attn(node_attn, edge_index):
    src_attn = node_attn[edge_index[0]]
    dst_attn = node_attn[edge_index[1]]
    edge_attn = src_attn * dst_attn
    return edge_attn


import math
import torch
import torch.nn as nn
from torch import Tensor
from functools import partial
from collections import Counter
import torch.nn.functional as F
from .shapley import gnn_score, mc_shapley, l_shapley, mc_l_shapley, NC_mc_l_shapley
import networkx as nx
from typing import Callable, Optional, Tuple
from torch_geometric.utils import to_networkx
from torch_geometric.utils.num_nodes import maybe_num_nodes
from torch_geometric.nn import MessagePassing
from torch_geometric.data import Data, Batch
from ..base import BaseRandom

def compute_scores(score_func, children):
    results = []
    not_good = 0
    label_nodes = set(torch.where(children[0].data.node_label)[0].cpu().numpy())
    for child in children:
        if child.P == 0:
            score = score_func(child.coalition, child.data)
            score_func(child.coalition, child.data)
            if child.data.node_label.sum().item() != 0 and score < 0.5 and set(child.coalition).intersection(label_nodes) == label_nodes:
                not_good = 1
        else:
            score = child.P
        results.append(score)
    if max(results) < 0.5 and not_good:
        print([(set(child.coalition).intersection(label_nodes) == label_nodes) for child in children])
        print([score_func(child.coalition, child.data) for child in children])
    return results


def find_closest_node_result(results, max_nodes):
    """ return the highest reward tree_node with its subgraph is smaller than max_nodes """

    results = sorted(results, key=lambda x: len(x.coalition))

    result_node = results[0]
    for result_idx in range(len(results)):
        x = results[result_idx]
        if len(x.coalition) <= max_nodes and x.P > result_node.P:
            result_node = x
    return result_node


def find_closest_node_result_list(results, nodes_num_list):
    result_list = []
    for node_num in nodes_num_list:  # 10, 12 increase
        result_node = find_closest_node_result(results, node_num)
        results.remove(result_node)
        result_list.append(result_node)
    return result_list


def reward_func(reward_method, value_func,
                local_radius=4, sample_num=100,
                subgraph_building_method='split'):
    if reward_method.lower() == 'gnn_score':
        return partial(gnn_score,
                       value_func=value_func,
                       subgraph_building_method=subgraph_building_method)

    elif reward_method.lower() == 'mc_shapley':
        return partial(mc_shapley,
                       value_func=value_func,
                       subgraph_building_method=subgraph_building_method,
                       sample_num=sample_num)

    elif reward_method.lower() == 'l_shapley':
        return partial(l_shapley,
                       local_raduis=local_radius,
                       value_func=value_func,
                       subgraph_building_method=subgraph_building_method)

    elif reward_method.lower() == 'mc_l_shapley':
        return partial(mc_l_shapley,
                       local_raduis=local_radius,
                       value_func=value_func,
                       subgraph_building_method=subgraph_building_method,
                       sample_num=sample_num)
    else:
        raise NotImplementedError


class MCTSNode(object):

    def __init__(self, coalition: list, data: Data, ori_graph: nx.Graph,
                 c_puct: float = 10.0, W: float = 0, N: int = 0, P: float = 0,
                 mapping=None):
        self.data = data
        self.signal = 1
        self.coalition = coalition # Coalition of possible subsets of players
        self.ori_graph = ori_graph # Original input graph
        self.c_puct = c_puct # Hyperparameter in search algorithm
        self.children = [] # Children within MCTS tree
        self.W = W  # sum of node value
        self.N = N  # times of arrival
        self.P = P  # property score (reward)

        self.mapping = mapping # ADDED from OWEN

    def Q(self): # Average of W
        return self.W / self.N if self.N > 0 else 0

    def U(self, n): # Action selection criteria for MCTS
        return self.c_puct * self.P * math.sqrt(n) / (1 + self.N)

def k_hop_subgraph_with_default_whole_graph(
        edge_index, node_idx=None, num_hops=3, relabel_nodes=False,
        num_nodes=None, flow='source_to_target'):
    r"""Computes the :math:`k`-hop subgraph of :obj:`edge_index` around node
    :attr:`node_idx`.
    It returns (1) the nodes involved in the subgraph, (2) the filtered
    :obj:`edge_index` connectivity, (3) the mapping from node indices in
    :obj:`node_idx` to their new location, and (4) the edge mask indicating
    which edges were preserved.
    Args:
        node_idx (int, list, tuple or :obj:`torch.Tensor`): The central
            node(s).
        num_hops: (int): The number of hops :math:`k`.
        edge_index (LongTensor): The edge indices.
        relabel_nodes (bool, optional): If set to :obj:`True`, the resulting
            :obj:`edge_index` will be relabeled to hold consecutive indices
            starting from zero. (default: :obj:`False`)
        num_nodes (int, optional): The number of nodes, *i.e.*
            :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
        flow (string, optional): The flow direction of :math:`k`-hop
            aggregation (:obj:`"source_to_target"` or
            :obj:`"target_to_source"`). (default: :obj:`"source_to_target"`)
    :rtype: (:class:`LongTensor`, :class:`LongTensor`, :class:`LongTensor`,
             :class:`BoolTensor`)
    """

    num_nodes = maybe_num_nodes(edge_index, num_nodes)

    assert flow in ['source_to_target', 'target_to_source']
    if flow == 'target_to_source':
        row, col = edge_index
    else:
        col, row = edge_index  # edge_index 0 to 1, col: source, row: target

    node_mask = row.new_empty(num_nodes, dtype=torch.bool)
    edge_mask = row.new_empty(row.size(0), dtype=torch.bool)
    inv = None

    if node_idx is None:
        subsets = torch.tensor([0])
        cur_subsets = subsets
        while 1:
            node_mask.fill_(False)
            node_mask[subsets] = True
            torch.index_select(node_mask, 0, row, out=edge_mask)
            subsets = torch.cat([subsets, col[edge_mask]]).unique()
            if not cur_subsets.equal(subsets):
                cur_subsets = subsets
            else:
                subset = subsets
                break

    node_mask.fill_(False)
    node_mask[subset] = True
    edge_mask = node_mask[row] & node_mask[col]

    edge_index = edge_index[:, edge_mask]

    if relabel_nodes:
        node_idx = row.new_full((num_nodes,), -1)
        node_idx[subset] = torch.arange(subset.size(0), device=row.device)
        edge_index = node_idx[edge_index]

    return subset, edge_index, inv, edge_mask  # subset: key new node idx; value original node idx


class MCTS(object):
    r"""
    Monte Carlo Tree Search Method
    Args:
        X (:obj:`torch.Tensor`): Input node features
        edge_index (:obj:`torch.Tensor`): The edge indices.
        num_hops (:obj:`int`): The number of hops :math:`k`.
        n_rollout (:obj:`int`): The number of sequence to build the monte carlo tree.
        min_atoms (:obj:`int`): The number of atoms for the subgraph in the monte carlo tree leaf node.
        c_puct (:obj:`float`): The hyper-parameter to encourage exploration while searching.
        expand_atoms (:obj:`int`): The number of children to expand.
        high2low (:obj:`bool`): Whether to expand children tree node from high degree nodes to low degree nodes.
        node_idx (:obj:`int`): The target node index to extract the neighborhood.
        score_func (:obj:`Callable`): The reward function for tree node, such as mc_shapely and mc_l_shapely.
    """
    # score_threshold = 0.3
    def __init__(self, graph, num_hops: int, use_mcts=True, use_pruning=True,
                 n_rollout: int = 20, min_atoms: int = 3, c_puct: float = 10.0,
                 expand_atoms: int = 14, high2low: bool = False, score_func: Callable = None, score_threshold: float = 0.3):

        self.num_hops = num_hops
        self.data = graph
        self.graph = to_networkx(self.data, to_undirected=True)  # NETWORKX VERSION OF GRAPH
        self.data = Batch.from_data_list([self.data])
        self.num_nodes = self.graph.number_of_nodes()
        self.score_func = score_func
        self.n_rollout = n_rollout
        self.min_atoms = min_atoms
        self.c_puct = c_puct
        self.expand_atoms = expand_atoms
        self.high2low = high2low
        self.score_threshold = score_threshold
        self.use_mcts = use_mcts
        self.use_pruning = use_pruning

        inv_mapping = None

        # extract the sub-graph and change the node indices.

        self.root_coalition = sorted([node for node in range(self.num_nodes)])
        self.MCTSNodeClass = partial(MCTSNode, data=self.data, ori_graph=self.graph, c_puct=self.c_puct, mapping=inv_mapping)
        self.root = self.MCTSNodeClass(self.root_coalition)  # Root of tree
        self.state_map = {str(self.root.coalition): self.root}

    def set_score_func(self, score_func):
        self.score_func = score_func

    @staticmethod
    def __subgraph__(node_idx, x, edge_index, num_hops, **kwargs):
        num_nodes, num_edges = x.size(0), edge_index.size(1)
        subset, edge_index, _, edge_mask = k_hop_subgraph_with_default_whole_graph(
            edge_index, node_idx, num_hops, relabel_nodes=True, num_nodes=num_nodes)

        x = x[subset]
        for key, item in kwargs.items():
            if torch.is_tensor(item) and item.size(0) == num_nodes:
                item = item[subset]
            elif torch.is_tensor(item) and item.size(0) == num_edges:
                item = item[edge_mask]
            kwargs[key] = item

        return x, edge_index, subset, edge_mask, kwargs

    def mcts_rollout(self, tree_node):
        cur_graph_coalition = tree_node.coalition
        if len(cur_graph_coalition) <= self.min_atoms:
            return tree_node.P
        if tree_node.signal == -1 and self.use_pruning:
            return tree_node.P

        # Expand if this node has never been visited
        if len(tree_node.children) == 0:
            node_degree_list = list(self.graph.subgraph(cur_graph_coalition).degree)
            node_degree_list = sorted(node_degree_list, key=lambda x: x[1], reverse=self.high2low)
            all_nodes = [x[0] for x in node_degree_list]

            if len(all_nodes) < self.expand_atoms:
                expand_nodes = all_nodes
            else:
                expand_nodes = all_nodes[:self.expand_atoms]

            for each_node in expand_nodes:
                # for each node, pruning it and get the remaining sub-graph
                # here we check the resulting sub-graphs and only keep the largest one
                subgraph_coalition = [node for node in all_nodes if node != each_node]

                subgraphs = [self.graph.subgraph(c)
                             for c in nx.connected_components(self.graph.subgraph(subgraph_coalition))]
                main_sub = subgraphs[0]
                for sub in subgraphs:
                    if sub.number_of_nodes() > main_sub.number_of_nodes():
                        main_sub = sub

                new_graph_coalition = sorted(subgraph_coalition)

                # check the state map and merge the same sub-graph
                Find_same = False
                for old_graph_node in self.state_map.values():
                    if Counter(old_graph_node.coalition) == Counter(new_graph_coalition):
                        new_node = old_graph_node
                        Find_same = True

                if Find_same == False:
                    new_node = self.MCTSNodeClass(new_graph_coalition)
                    self.state_map[str(new_graph_coalition)] = new_node

                Find_same_child = False
                for cur_child in tree_node.children:
                    if Counter(cur_child.coalition) == Counter(new_graph_coalition):
                        Find_same_child = True

                if Find_same_child == False:
                    tree_node.children.append(new_node)


            scores = compute_scores(self.score_func, tree_node.children)

            # label_nodes = set(torch.where(self.data.node_label)[0].numpy())
            for child, score in zip(tree_node.children, scores):
                child.P = score
                if score < self.score_threshold and self.use_pruning:
                    child.signal = -1
                # if score < 0.5 and len(set(child.coalition).intersection(label_nodes)) > int(len(label_nodes)/2):
                    # print(f"Labeled Nodes: {label_nodes}, Subgraph: {set(child.coalition)}, Value: {score}")
                # elif score < self.score_threshold
                #     pass
                # elif score > self.score_threshold and set(child.coalition).isdisjoint(label_nodes) and score > 0.5:
                #     pass
                    # print(f"Labeled Nodes: {label_nodes}, Subgraph: {set(child.coalition)}, Value: {score}")

        sum_count = sum([c.N for c in tree_node.children])
        if self.use_mcts:
            selected_node = max(tree_node.children, key=lambda x: x.Q() + x.U(sum_count))
            v = self.mcts_rollout(selected_node)
            selected_node.W += v
            selected_node.N += 1
            return v
        else:
            assert self.use_mcts is False
            selected_node = max(tree_node.children, key=lambda x: x.P)
            # if selected_node.signal == -1:
            #     return selected_node.P
            v = self.mcts_rollout(selected_node)
            return v

    def mcts(self, verbose=True):

        if verbose:
            print(f"The nodes in graph is {self.graph.number_of_nodes()}")
        # for rollout_idx in range(self.n_rollout):
        self.mcts_rollout(self.root)
        # print('no_rollout')
            # if verbose:
            #     print(f"At the {rollout_idx} rollout, {len(self.state_map)} states that have been explored.")

        explanations = [node for _, node in self.state_map.items()]
        explanations = sorted(explanations, key=lambda x: x.P, reverse=True)
        # Sorts explanations based on P value (i.e. Score(.,.,.) function in MCTS)
        return explanations


class SubgraphX(BaseRandom):
    r"""
    Code adapted from Dive into Graphs (DIG)
    Code: https://github.com/divelab/DIG
    The implementation of paper
    `On Explainability of Graph Neural Networks via Subgraph Explorations <https://arxiv.org/abs/2102.05152>`_.
    Args:
        model (:obj:`torch.nn.Module`): The target model prepared to explain
        num_hops(:obj:`int`, :obj:`None`): The number of hops to extract neighborhood of target node
          (default: :obj:`None`)
        rollout(:obj:`int`): Number of iteration to get the prediction
        min_atoms(:obj:`int`): Number of atoms of the leaf node in search tree
        c_puct(:obj:`float`): The hyperparameter which encourages the exploration
        expand_atoms(:obj:`int`): The number of atoms to expand
          when extend the child nodes in the search tree
        high2low(:obj:`bool`): Whether to expand children nodes from high degree to low degree when
          extend the child nodes in the search tree (default: :obj:`False`)
        local_radius(:obj:`int`): Number of local radius to calculate :obj:`l_shapley`, :obj:`mc_l_shapley`
        sample_num(:obj:`int`): Sampling time of monte carlo sampling approximation for
          :obj:`mc_shapley`, :obj:`mc_l_shapley` (default: :obj:`mc_l_shapley`)
        reward_method(:obj:`str`): The command string to select the
        subgraph_building_method(:obj:`str`): The command string for different subgraph building method,
          such as :obj:`zero_filling`, :obj:`split` (default: :obj:`zero_filling`)
    Example:
        >>> # For graph classification task
        >>> subgraphx = SubgraphX(model=model, num_classes=2)
    """
    rollout: int = 20
    min_atoms: int = 3
    c_puct: float = 10.0
    expand_atoms = 15
    local_radius = 4
    sample_num = 100
    reward_method = 'mc_l_shapley'

    def __init__(self, clf, criterion, config):
        super().__init__()
        num_hops = None
        self.name = 'subgraphx'
        self.clf = clf
        self.device = next(self.parameters()).device
        self.sparsity_set = config['sparsity_set']
        # extractor wasn't used
        # self.extractor = extractor
        self.num_hops = self.update_num_hops(num_hops)
        self.criterion = criterion
        self.score_threshold = config['score_threshold']
        self.use_mcts = config['use_mcts']
        self.use_pruning = config['use_pruning']
        self.high2low = config['high2low']  # False
        self.subgraph_building_method = config['subgraph_building_method']  # "zero_filling"
        # # mcts hyper-parameters
        # self.rollout = rollout
        # self.min_atoms = min_atoms  # N_{min}
        # self.c_puct = c_puct
        # self.expand_atoms = expand_atoms
        # self.high2low = high2low
        #
        # # reward function hyper-parameters
        # self.local_radius = local_radius
        # self.sample_num = sample_num
        # self.reward_method = reward_method
        # self.subgraph_building_method = subgraph_building_method

    def forward_pass(self, data, epoch, do_sampling, **kwargs):
        x_level = 'geometric'
        clf_logits = self.clf(data)
        batch_imp = []
        for graph in data.to_data_list():
            imp = self.get_explanation_graph(graph, x_level, self.sparsity_set)
            batch_imp += [imp]
        return -1, {}, clf_logits, torch.cat(batch_imp)

    def update_num_hops(self, num_hops):
        if num_hops is not None:
            return num_hops

        k = 0
        for module in self.clf.modules():
            if isinstance(module, MessagePassing):
                k += 1
        return k

    def get_reward_func(self, value_func):
        return reward_func(reward_method=self.reward_method,
                           value_func=value_func,
                           local_radius=self.local_radius,
                           sample_num=self.sample_num,
                           subgraph_building_method=self.subgraph_building_method)

    def get_mcts_class(self, graph, score_func: Callable = None):
        # only for graph classification
        return MCTS(graph, use_mcts=self.use_mcts,
                    use_pruning=self.use_pruning,
                    score_func=score_func,
                    num_hops=self.num_hops,
                    n_rollout=self.rollout,
                    min_atoms=self.min_atoms,
                    c_puct=self.c_puct,
                    expand_atoms=self.expand_atoms,
                    high2low=self.high2low,
                    score_threshold=self.score_threshold)

    def get_explanation_graph(self,
                              graph,
                              x_level: str,
                              sparsity_set=None,
                              max_nodes: int = 14,
                              ):
        '''
        Get explanation for a whole graph prediction.
        Args:
            x (torch.Tensor): Input features for every node in the graph.
            edge_index (torch.Tensor): Edge index for entire input graph.
            label (int, optional): Label for which to assume as a prediction from
                the model when generating an explanation. If `None`, this argument
                is set to the prediction directly from the model. (default: :obj:`None`)
            max_nodes (int, optional): Maximum number of nodes to include in the subgraph
                generated from the explanation. (default: :obj:`14`)
            forward_kwargs (dict, optional): Additional arguments to model.forward
                beyond x and edge_index. Must be keyed on argument name.
                (default: :obj:`{}`)
        :rtype: :class:`Explanation`
        Returns:
            exp (dict):
                exp['feature_imp'] is `None` because no feature explanations are generated.
                exp['node_imp'] (torch.Tensor, (n,)): Node mask of size `(n,)` where `n`
                    is number of nodes in the entire graph described by `edge_index`.
                    Type is `torch.bool`, with `True` indices corresponding to nodes
                    included in the subgraph.
                exp['edge_imp'] (torch.Tensor, (e,)): Edge mask of size `(e,)` where `e`
                    is number of edges in the entire graph described by `edge_index`.
                    Type is `torch.bool`, with `True` indices corresponding to edges
                    included in the subgraph.
        '''
        if not sparsity_set:
            sparsity_set = {0.3, 0.5, 0.7}
        sub_nodes_list = []
        for sparsity in sorted(sparsity_set, reverse=True):
            sub_nodes_list += [int(graph.num_nodes * (1 - sparsity))]
        edge_index = graph.edge_index

        label = graph.y.item()
        # self.clf.eval()
        # pred = self.clf(graph).sigmoid()
        # label = int(pred.item())
        if label != graph.y.item():
            print(f"Gnd: {graph.y.item()}, Pred: {label}, SigNodes: {set(torch.where(graph.node_label)[0].cpu().numpy())}")
        # collect all the class index
        # logits = self.model(graph)
        # probs = F.softmax(logits, dim=-1)
        # probs = probs.squeeze()
        #
        # prediction = probs.argmax(-1)

        value_func = self._prob_score_func_graph(target_class=label)

        payoff_func = self.get_reward_func(value_func)
        self.mcts_state_map = self.get_mcts_class(graph, score_func=payoff_func)
        results = self.mcts_state_map.mcts(verbose=False)

        # best_result = find_closest_node_result(results, max_nodes=max_nodes)
        # node_mask, edge_mask = self.__parse_results(best_result, edge_index)

        best_results = find_closest_node_result_list(results, sub_nodes_list)
        node_mask, edge_mask = self.__parse_results_list(best_results, edge_index)


        node_imp = node_mask.float()
        edge_imp = edge_mask.float()

        # exp.node_imp = node_mask
        # exp.edge_imp = edge_mask

        # return {'feature_imp': None, 'node_imp': node_mask, 'edge_imp': edge_mask}
        return edge_imp if hasattr(graph, 'edge_label') else node_imp

    def _prob_score_func_graph(self, target_class):
        """
        Get a function that computes the predicted probability that the input graphs
        are classified as target classes.
        Args:
            target_class (int): the targeted class of the graph
        Returns:
            get_prob_score (callable): the probability score function
        """
        def get_prob_score(graph):
            prob = self.clf(graph).sigmoid()
            score = prob if target_class == 1 else 1-prob
            # score = prob[:, target_class]
            return score

        return get_prob_score

    def __parse_results_list(self, subgraph_list, edge_index):
        num_nodes = maybe_num_nodes(edge_index)
        node_mask = torch.zeros(num_nodes, device=edge_index.device)
        imp_list = [0.1 * i for i in range(11-len(subgraph_list), 11, 1)]
        # the size of subgraph is increasing, so small subgraph has higher imp
        for subgraph, imp in zip(reversed(subgraph_list), imp_list):
            n_mask, e_mask = self.__parse_results(subgraph, edge_index, imp=imp)
            # node_mask = torch.where(node_mask == 0, n_mask, node_mask)
            node_mask = torch.where(node_mask < n_mask, n_mask, node_mask)
        edge_mask = node_mask[edge_index[0]] * node_mask[edge_index[1]]
        return node_mask, edge_mask

    def __parse_results(self, best_subgraph, edge_index, imp=1.0):
        # Function strongly based on torch_geometric.utils.subgraph function
        # Citation: https://pytorch-geometric.readthedocs.io/en/latest/_modules/torch_geometric/utils/subgraph.html#subgraph

        # Get mapping
        map = best_subgraph.mapping

        all_nodes = torch.unique(edge_index)

        subgraph_nodes = torch.tensor([map[c] for c in best_subgraph.coalition], dtype=torch.long) if map is not None \
            else torch.tensor(best_subgraph.coalition, dtype=torch.long)

        # Create node mask:
        node_mask = torch.zeros(max(all_nodes)+1, device=edge_index.device)
        node_mask[subgraph_nodes] = imp

        # Create edge_index mask
        num_nodes = maybe_num_nodes(edge_index)
        n_mask = torch.zeros(num_nodes, dtype=torch.bool, device=edge_index.device)
        n_mask[subgraph_nodes] = imp

        edge_mask = n_mask[edge_index[0]] & n_mask[edge_index[1]]
        return node_mask, edge_mask